from collections import defaultdict
from typing import Any, List, Set

from gambatools.automaton import Automaton
from gambatools.dfa import DFA
from gambatools.nfa import NFA
from gambatools.pda import PDA
from gambatools.tm import TM
from gambatools.text_utility import read_text, write_text


SIGMA_HEADER = r'''<script src="https://cdn.jsdelivr.net/npm/sigma@1.2.1/src/sigma.core.js"></script>
<script src="https://cdn.jsdelivr.net/npm/sigma@1.2.1/src/conrad.js"></script>
<script src="https://cdn.jsdelivr.net/npm/sigma@1.2.1/src/utils/sigma.utils.js"></script>
<script src="https://cdn.jsdelivr.net/npm/sigma@1.2.1/src/utils/sigma.polyfills.js"></script>
<script src="https://cdn.jsdelivr.net/npm/sigma@1.2.1/src/sigma.settings.js"></script>
<script src="https://cdn.jsdelivr.net/npm/sigma@1.2.1/src/classes/sigma.classes.dispatcher.js"></script>
<script src="https://cdn.jsdelivr.net/npm/sigma@1.2.1/src/classes/sigma.classes.configurable.js"></script>
<script src="https://cdn.jsdelivr.net/npm/sigma@1.2.1/src/classes/sigma.classes.graph.js"></script>
<script src="https://cdn.jsdelivr.net/npm/sigma@1.2.1/src/classes/sigma.classes.camera.js"></script>
<script src="https://cdn.jsdelivr.net/npm/sigma@1.2.1/src/classes/sigma.classes.quad.js"></script>
<script src="https://cdn.jsdelivr.net/npm/sigma@1.2.1/src/classes/sigma.classes.edgequad.js"></script>
<script src="https://cdn.jsdelivr.net/npm/sigma@1.2.1/src/captors/sigma.captors.mouse.js"></script>
<script src="https://cdn.jsdelivr.net/npm/sigma@1.2.1/src/captors/sigma.captors.touch.js"></script>
<script src="https://cdn.jsdelivr.net/npm/sigma@1.2.1/src/renderers/sigma.renderers.canvas.js"></script>
<script src="https://cdn.jsdelivr.net/npm/sigma@1.2.1/src/renderers/sigma.renderers.webgl.js"></script>
<script src="https://cdn.jsdelivr.net/npm/sigma@1.2.1/src/renderers/sigma.renderers.svg.js"></script>
<script src="https://cdn.jsdelivr.net/npm/sigma@1.2.1/src/renderers/sigma.renderers.def.js"></script>
<script src="https://cdn.jsdelivr.net/npm/sigma@1.2.1/src/renderers/canvas/sigma.canvas.labels.def.js"></script>
<script src="https://cdn.jsdelivr.net/npm/sigma@1.2.1/src/renderers/canvas/sigma.canvas.hovers.def.js"></script>
<script src="https://cdn.jsdelivr.net/npm/sigma@1.2.1/src/renderers/canvas/sigma.canvas.nodes.def.js"></script>
<script src="https://cdn.jsdelivr.net/npm/sigma@1.2.1/src/renderers/canvas/sigma.canvas.edges.def.js"></script>
<script src="https://cdn.jsdelivr.net/npm/sigma@1.2.1/src/renderers/canvas/sigma.canvas.edges.curve.js"></script>
<script src="https://cdn.jsdelivr.net/npm/sigma@1.2.1/src/renderers/canvas/sigma.canvas.edges.arrow.js"></script>
<script src="https://cdn.jsdelivr.net/npm/sigma@1.2.1/src/renderers/canvas/sigma.canvas.edges.curvedArrow.js"></script>
<script src="https://cdn.jsdelivr.net/npm/sigma@1.2.1/src/renderers/canvas/sigma.canvas.edgehovers.def.js"></script>
<script src="https://cdn.jsdelivr.net/npm/sigma@1.2.1/src/renderers/canvas/sigma.canvas.edgehovers.curve.js"></script>
<script src="https://cdn.jsdelivr.net/npm/sigma@1.2.1/src/renderers/canvas/sigma.canvas.edgehovers.arrow.js"></script>
<script src="https://cdn.jsdelivr.net/npm/sigma@1.2.1/src/renderers/canvas/sigma.canvas.edgehovers.curvedArrow.js"></script>
<script src="https://cdn.jsdelivr.net/npm/sigma@1.2.1/src/renderers/canvas/sigma.canvas.extremities.def.js"></script>
<script src="https://cdn.jsdelivr.net/npm/sigma@1.2.1/src/middlewares/sigma.middlewares.rescale.js"></script>
<script src="https://cdn.jsdelivr.net/npm/sigma@1.2.1/src/middlewares/sigma.middlewares.copy.js"></script>
<script src="https://cdn.jsdelivr.net/npm/sigma@1.2.1/src/misc/sigma.misc.animation.js"></script>
<script src="https://cdn.jsdelivr.net/npm/sigma@1.2.1/src/misc/sigma.misc.bindEvents.js"></script>
<script src="https://cdn.jsdelivr.net/npm/sigma@1.2.1/src/misc/sigma.misc.bindDOMEvents.js"></script>
<script src="https://cdn.jsdelivr.net/npm/sigma@1.2.1/src/misc/sigma.misc.drawHovers.js"></script>
<!-- Sigma plugins -->
<script src="https://cdn.jsdelivr.net/npm/sigma@1.2.1/plugins/sigma.layout.forceAtlas2/supervisor.js"></script>
<script src="https://cdn.jsdelivr.net/npm/sigma@1.2.1/plugins/sigma.layout.forceAtlas2/worker.js"></script>
<script src="https://cdn.jsdelivr.net/npm/sigma@1.2.1/plugins/sigma.renderers.edgeLabels/settings.js"></script>
<script src="https://cdn.jsdelivr.net/npm/sigma@1.2.1/plugins/sigma.renderers.edgeLabels/sigma.canvas.edges.labels.def.js"></script>
<script src="https://cdn.jsdelivr.net/npm/sigma@1.2.1/plugins/sigma.renderers.edgeLabels/sigma.canvas.edges.labels.curve.js"></script>
<script src="https://cdn.jsdelivr.net/npm/sigma@1.2.1/plugins/sigma.renderers.edgeLabels/sigma.canvas.edges.labels.curvedArrow.js"></script>
'''

SIGMA_TEMPLATE = r'''
<div id="<CONTAINER>" style="<STYLE>"></div>
<script>
// function for drawing a state
sigma.canvas.nodes.state = function(node, context, settings)
{
  var prefix = settings('prefix') || '';

  context.beginPath();
  context.arc(
    node[prefix + 'x'],
    node[prefix + 'y'],
    node[prefix + 'size'],
    0,
    Math.PI * 2,
    true
  );
  context.fillStyle = node.initial ? 'lightgreen' : 'white';
  context.fill();
  context.strokeStyle = 'black';
  context.stroke();

  if (node.accept)
  {
    var d = 4;
    context.arc(
      node[prefix + 'x'],
      node[prefix + 'y'],
      node[prefix + 'size'],
      0,
      Math.PI * 2,
      true
    );
    context.stroke();
  }
  context.closePath();
};

// function for drawing a state label
sigma.canvas.labels.state = function(node, context, settings)
{
  var fontSize,
      prefix = settings('prefix') || '',
      size = node[prefix + 'size'];

  if (size < settings('labelThreshold'))
    return;

  if (!node.label || typeof node.label !== 'string')
    return;

  fontSize = (settings('labelSize') === 'fixed') ? settings('defaultLabelSize') : settings('labelSizeRatio') * size;

  context.font = (settings('fontStyle') ? settings('fontStyle') + ' ' : '') + fontSize + 'px ' + settings('font');
  context.fillStyle = 'black';

  var labelWidth = context.measureText(node.label).width;
  var labelPlacementX = Math.round(node[prefix + 'x'] - labelWidth / 2);
  var labelPlacementY = Math.round(node[prefix + 'y'] + fontSize / 3);
  context.fillText(node.label, labelPlacementX, labelPlacementY);
};

var s = new sigma(
  {
    renderer: {
      container: document.getElementById('<CONTAINER>'),
      type: 'canvas'
    },
    settings: {
      edgeLabelSize: 'fixed',
      minArrowSize: 15,
      minNodeSize: 18,
      maxNodeSize: 30,
      defaultLabelSize: 14,
      defaultEdgeLabelSize: 16,
    }
  }
);

var graph =
{
  nodes:
  [
    <NODES>
  ],
  edges:
  [
    <EDGES>
  ]
};

s.graph.read(graph);
s.refresh();
s.startForceAtlas2();
window.setTimeout(function() {s.killForceAtlas2()}, 3000);
</script>
'''


HTML_TEMPLATE = r'''<!DOCTYPE html>
<html>
<head>
<meta charset="utf-8"/>
<title>Automaton Visualization</title>
{}
<style>
  html {{ height:100%; }}
  body {{ height:100%; }}
</style>
</head>
<body>
{}
</body>
</html>
'''


def default_html_settings():
    return {'container': 'sigma-container', 'style': 'width:100%; height:100%; background-color:#E1E1E1'}


# {id: 0, label: 'q0', x: Math.random(), y: Math.random(), size: 1, type: 'state', initial: true},
def make_node(id: int, label: str, accept: bool = False, initial: bool = False) -> str:
    attributes = {
        'id': id,
        'label': "'{}'".format(label),
        'x': 'Math.random()',
        'y': 'Math.random()',
        'size': 1,
        'type': "'state'",
    }
    if accept:
        attributes['accept'] = 'true'
    if initial:
        attributes['initial'] = 'true'
    items = ["{}: {}".format(key, value) for (key, value) in attributes.items()]
    return '{{ {} }}'.format(', '.join(items))


#    { id: 0, label: 'a', source: '0', target: '1', color: '#00000', type: 'arrow' },
#    { id: 1, label: 'b', source: '1', target: '0', color: '#00000', type: 'curvedArrow' },
def make_edge(id: int, label: str, source: int, target: int, color: str, type_: str) -> str:
    attributes = {
        'id': id,
        'label': "'{}'".format(label),
        'source': "{}".format(source),
        'target': "{}".format(target),
        'color': "'{}'".format(color),
        'type': "'{}'".format(type_),
    }
    items = ["{}: {}".format(key, value) for (key, value) in attributes.items()]
    return '{{ {} }}'.format(', '.join(items))


def make_html(nodes: List[str], edges: List[str], html_settings=default_html_settings()) -> str:
    text = SIGMA_TEMPLATE
    text = text.replace('<CONTAINER>', html_settings['container'])
    text = text.replace('<STYLE>', html_settings['style'])
    text = text.replace('<NODES>', ',\n    '.join(nodes))
    text = text.replace('<EDGES>', ',\n    '.join(edges))
    return text


def compute_nodes(states: Set[Any], initial_states: Set[Any], final_states: Set[Any] = frozenset([])):
    # compute nodes
    nodes = []
    node_map = {}
    for i, q in enumerate(states):
        initial = q in initial_states
        accept = q in final_states
        nodes.append(make_node(i, q, accept, initial))
        node_map[q] = i
    return nodes, node_map


def automaton_to_sigma(A: Automaton, html_settings=default_html_settings()) -> str:
    initial_states = A.initial_states
    states = A.states
    transitions = A.transitions
    final_states = A.final_states

    nodes, node_map = compute_nodes(states, initial_states, final_states)

    # compute edges
    edge_map = defaultdict(lambda: [])
    for (p, a, q) in transitions:
        edge_map[p, q].append(a)
    edges = []
    for i, ((p, q), labels) in enumerate(edge_map.items()):
        type_ = 'curvedArrow' if (q, p) in edge_map else 'arrow'
        label = ','.join(labels)
        label = label.replace('_', 'ε')
        edges.append(make_edge(i, label, node_map[p], node_map[q], 'black', type_))
    return make_html(nodes, edges, html_settings)


def dfa_to_sigma(D: DFA, html_settings=default_html_settings()) -> str:
    q0 = D.q0
    Q = D.Q
    F = D.F
    delta = D.delta

    nodes, node_map = compute_nodes(Q, {q0}, F)

    # compute edges
    edge_map = defaultdict(lambda: [])
    for (q, a), q1 in delta.items():
        edge_map[q, q1].append(a)
    edges = []
    for i, ((q, q1), labels) in enumerate(edge_map.items()):
        type_ = 'curvedArrow' if (q1, q) in edge_map else 'arrow'
        label = ' '.join(labels)
        edges.append(make_edge(i, label, node_map[q], node_map[q1], 'black', type_))
    return make_html(nodes, edges, html_settings)


def nfa_to_sigma(N: NFA, html_settings=default_html_settings()) -> str:
    q0 = N.q0
    Q = N.Q
    F = N.F
    delta = N.delta
    epsilon = N.epsilon

    nodes, node_map = compute_nodes(Q, {q0}, F)

    # compute edges
    edge_map = defaultdict(lambda: [])
    for (q, a), Q1 in delta.items():
        a = 'ε' if a == epsilon else a
        for q1 in Q1:
            edge_map[q, q1].append(a)
    edges = []
    for i, ((q, q1), labels) in enumerate(edge_map.items()):
        type_ = 'curvedArrow' if (q1, q) in edge_map else 'arrow'
        label = ','.join(labels)
        edges.append(make_edge(i, label, node_map[q], node_map[q1], 'black', type_))
    return make_html(nodes, edges, html_settings)


def pda_to_sigma(P: PDA, html_settings=default_html_settings()) -> str:
    q0 = P.q0
    Q = P.Q
    F = P.F
    delta = P.delta
    epsilon = P.epsilon

    nodes, node_map = compute_nodes(Q, {q0}, F)

    # compute edges
    edge_map = defaultdict(lambda: [])
    for (p, a, u), Q1 in delta.items():
        for (q, v) in Q1:
            a = 'ε' if a == epsilon else a
            u = 'ε' if u == epsilon else u
            v = 'ε' if v == epsilon else v
            edge_map[p, q].append('{},{}→{}'.format(a, u, v))
    edges = []
    for i, ((p, q), labels) in enumerate(edge_map.items()):
        type_ = 'curvedArrow' if (q, p) in edge_map else 'arrow'
        label = ','.join(labels)
        edges.append(make_edge(i, label, node_map[p], node_map[q], 'black', type_))
    return make_html(nodes, edges, html_settings)


def tm_to_sigma(T: TM, html_settings=default_html_settings()) -> str:
    q0 = T.q0
    Q = T.Q
    delta = T.delta
    blank = T.blank

    nodes, node_map = compute_nodes(Q, {q0})

    # compute edges
    edge_map = defaultdict(lambda: [])
    for (p, a), (q, b, d) in delta.items():
        a = '□' if a == blank else a
        b = '□' if b == blank else b
        edge_map[p, q].append('{}→{},{}'.format(a, b, d))
    edges = []
    for i, ((p, q), labels) in enumerate(edge_map.items()):
        type_ = 'curvedArrow' if (q, p) in edge_map else 'arrow'
        label = ' '.join(labels)
        edges.append(make_edge(i, label, node_map[p], node_map[q], 'black', type_))
    return make_html(nodes, edges, html_settings)


def draw_dfa(D: DFA, filename: str) -> None:
    write_text(filename, HTML_TEMPLATE.format(SIGMA_HEADER, dfa_to_sigma(D)))


def draw_nfa(N: NFA, filename: str) -> None:
    write_text(filename, HTML_TEMPLATE.format(SIGMA_HEADER, nfa_to_sigma(N)))
    write_text(filename, SIGMA_HEADER + nfa_to_sigma(N))


def draw_pda(P: PDA, filename: str) -> None:
    write_text(filename, HTML_TEMPLATE.format(SIGMA_HEADER, pda_to_sigma(P)))


def draw_tm(T: TM, filename: str) -> None:
    write_text(filename, HTML_TEMPLATE.format(SIGMA_HEADER, tm_to_sigma(T)))



#  (C) Copyright Wieger Wesselink 2020. Distributed under the GPL-3.0-or-later
#  Software License, (See accompanying file LICENSE or copy at
#  https://www.gnu.org/licenses/gpl-3.0.txt)

from typing import Callable, Optional, Tuple, List, Union, Set
import re
import graphviz
from tabulate import tabulate

from gambatools.dfa import DFA
from gambatools.language_algorithms import parse_word_list
from gambatools.nfa import NFA
from gambatools.pda import PDA
from gambatools.tm import TM
from gambatools.cfg import CFG
from gambatools.regexp import Regexp
from gambatools.dfa_algorithms import State, Symbol, parse_dfa, dfa_accepts_word
from gambatools.automaton_algorithms import parse_automaton, state_set_regex, default_state_label_regex, \
    state_product_regex
from gambatools.automaton_io import automaton_to_dot
from gambatools.cfg_algorithms import cfg_accepts_word, parse_simple_cfg, cfg_words_up_to_n
from gambatools.dfa_algorithms import dfa_words_up_to_n, dfa_simulate_word
from gambatools.nfa_algorithms import nfa_accepts_word, nfa_simulate_word, nfa_words_up_to_n, parse_nfa
from gambatools.pda_algorithms import pda_simulate_word, pda_accepts_word, parse_pda, pda_words_up_to_n
from gambatools.tm_algorithms import tm_accepts_word, tm_simulate_word, parse_tm, tm_words_up_to_n
from gambatools.printing import print_words
from gambatools.text_utility import read_utf8_text
from gambatools.language_generator import check_equal_languages, compare_languages, generate_language
from gambatools.regexp_algorithms import regexp_words_up_to_n, regexp_accepts_word
from gambatools.regexp_simple_parser import parse_simple_regexp


def check_word_in_language(word: str, Sigma: Set[str]) -> None:
    for w in word:
        if not w in Sigma:
            raise RuntimeError('the symbol {} is not in the alphabet {{{}}}'.format(w, Sigma))


def check_automaton_syntax(text: str, parse: Callable) -> None:
    try:
        parse(text)
        print('OK')
    except Exception as e:
        print('Error: {}'.format(e))


def check_dfa_syntax(text: str):
    check_automaton_syntax(text, parse_dfa)


def check_nfa_syntax(text: str):
    check_automaton_syntax(text, parse_nfa)


def check_pda_syntax(text: str):
    check_automaton_syntax(text, parse_pda)


def check_tm_syntax(text: str):
    check_automaton_syntax(text, parse_tm)


def show_automaton(text: str, check: Optional[Callable] = None, state_regex=default_state_label_regex()) -> graphviz.Digraph:
    try:
        if check:
            check(text)
        A = parse_automaton(text, state_regex=state_regex)
        return automaton_to_dot(A)
    except Exception as e:
        print('Error: {}'.format(e))


def show(text: str) -> graphviz.Digraph:
    return show_automaton(text)


def show_product(text: str) -> graphviz.Digraph:
    return show_automaton(text, state_regex=state_product_regex())


def show_nfa2dfa(text: str) -> graphviz.Digraph:
    return show_automaton(text, state_regex=state_set_regex())


def show_dfa(text: str) -> graphviz.Digraph:
    return show_automaton(text, parse_dfa)


def show_nfa(text: str) -> graphviz.Digraph:
    return show_automaton(text, parse_nfa)


def show_pda(text: str) -> graphviz.Digraph:
    return show_automaton(text, parse_pda)


def show_tm(text: str) -> graphviz.Digraph:
    return show_automaton(text, parse_tm)


def language_parser(filename: str) -> Callable:
    if filename.endswith('.dfa'):
        return parse_dfa
    elif filename.endswith('.nfa'):
        return parse_nfa
    elif filename.endswith('.pda'):
        return parse_pda
    elif filename.endswith('.tm'):
        return parse_tm
    elif filename.endswith('.cfg'):
        return parse_simple_cfg
    elif filename.endswith('.regexp'):
        return parse_simple_regexp
    raise RuntimeError('Error: unknown extension in file {}'.format(filename))


def check_language_from_file(text: str, text_parser: Callable, answerfile: str, length: int = 5) -> None:
    try:
        A1 = text_parser(text)
        A2 = language_parser(answerfile)(read_utf8_text(answerfile))
        feedback = check_equal_languages(A1, A2, length)
        print_feedback(feedback)
    except Exception as e:
        print('Error: {}'.format(e))


def check_dfa_language_from_file(text: str, filename: str, length: int = 5):
    check_language_from_file(text, parse_dfa, filename, length)


def check_nfa_language_from_file(text: str, filename: str, length: int = 5):
    check_language_from_file(text, parse_nfa, filename, length)


def check_pda_language_from_file(text: str, filename: str, length: int = 5):
    check_language_from_file(text, parse_pda, filename, length)


def check_tm_language_from_file(text: str, filename: str, length: int = 5):
    check_language_from_file(text, parse_tm, filename, length)


def check_cfg_language_from_file(text: str, filename: str, length: int = 5):
    check_language_from_file(text, parse_simple_cfg, filename, length)


def check_regexp_language_from_file(text: str, filename: str, length: int = 5):
    check_language_from_file(text, parse_simple_regexp, filename, length)


def check_max_states(A: Union[DFA, NFA, PDA, TM], max_states: int) -> List[str]:
    if 0 < max_states < len(A.Q):
        return ['Error: the maximum number of states ({}) is exceeded'.format(max_states)]
    return []


def print_feedback(feedback: List[str]) -> None:
    if feedback:
        for msg in feedback:
            print(msg)
    else:
        print('OK')


def check_language_from_words(text: str, parser: Callable, word_list: str, length: int, max_states: int = 0):
    try:
        A = parser(text)
        feedback = check_max_states(A, max_states)
        A_words = generate_language(A, length)
        words = parse_word_list(word_list)
        feedback.extend(compare_languages(A_words, words))
        print_feedback(feedback)
    except Exception as e:
        print('Error: {}'.format(e))


def check_dfa_language_from_words(text: str, word_list: str, length: int, max_states: int = 0):
    check_language_from_words(text, parse_dfa, word_list, length, max_states)


def check_nfa_language_from_words(text: str, word_list: str, length: int, max_states: int = 0):
    check_language_from_words(text, parse_nfa, word_list, length, max_states)


def check_pda_language_from_words(text: str, word_list: str, length: int, max_states: int = 0):
    check_language_from_words(text, parse_pda, word_list, length, max_states)


def check_tm_language_from_words(text: str, word_list: str, length: int, max_states: int = 0):
    check_language_from_words(text, parse_tm, word_list, length, max_states)


def check_cfg_language_from_words(text: str, word_list: str, length: int):
    check_language_from_words(text, parse_simple_cfg, word_list, length)


def check_regexp_language_from_words(text: str, word_list: str, length: int):
    check_language_from_words(text, parse_simple_regexp, word_list, length)


def simulate_dfa(text: str, word: str) -> None:
    from IPython.core.display import display, HTML

    try:
        D = parse_dfa(text)
        check_word_in_language(word, D.Sigma)
        table = dfa_simulate_word(D, word)
        html = tabulate(table, headers=["State", "Word"], tablefmt='html')
        display(HTML(html))
    except Exception as e:
        print('Error: {}'.format(e))


def simulate_nfa(text: str, word: str) -> None:
    from IPython.core.display import display, HTML

    def make_row(row: Tuple[State, str]):
        q, word = row
        word = word if word else 'ε'
        return q, word

    try:
        N = parse_nfa(text)
        check_word_in_language(word, N.Sigma)
        if nfa_accepts_word(N, word):
            table = nfa_simulate_word(N, word)
            table = [make_row(row) for row in table]
            html = tabulate(table, headers=["State", "Word"], tablefmt='html')
            display(HTML(html))
        else:
            print('word {} can not be simulated, since it is not in the language'.format(word))
    except Exception as e:
        print('Error: {}'.format(e))


def simulate_pda(text: str, word: str) -> None:
    from IPython.core.display import display, HTML

    def make_row(row: Tuple[State, str, List[Symbol]], epsilon: str):
        q, word, stack = row
        stack = [t if t != epsilon else 'ε' for t in stack]
        word = word if word else 'ε'
        stack = re.sub(epsilon, 'ε', ''.join(stack))
        return q, word, stack

    try:
        P = parse_pda(text)
        check_word_in_language(word, P.Sigma)
        if pda_accepts_word(P, word):
            table = pda_simulate_word(P, word)
            table = [make_row(row, P.epsilon) for row in table]
            html = tabulate(table, headers=["State", "Word", "Stack"], tablefmt='html')
            display(HTML(html))
        else:
            print('Warning: {} can not be simulated, since it is not in the language'.format(word))
    except Exception as e:
        print('Error: {}'.format(e))


def simulate_tm(text: str, word: str) -> None:
    from IPython.core.display import display, HTML

    def make_row(row: Tuple[State, List[str], int], blank: str) -> Tuple[str, str]:
        q, tape, pos = row
        tape = tape[:]
        while pos < len(tape) - 1 and tape[-1] == blank:
            tape.pop()
        tape = [t if t != blank else '□' for t in tape]
        tape[pos] = 'UNDERLINE_LEFT{}UNDERLINE_RIGHT'.format(tape[pos])  # TODO: find a cleaner solution for this
        return str(q), ''.join(tape)

    try:
        T = parse_tm(text)
        check_word_in_language(word, T.Sigma)
        table = tm_simulate_word(T, word)
        table = [make_row(row, T.blank) for row in table]
        html = tabulate(table, headers=["State", "Tape", "Head"], tablefmt='html')
        html = html.replace('UNDERLINE_LEFT', '<u>')
        html = html.replace('UNDERLINE_RIGHT', '</u>')
        display(HTML(html))
    except Exception as e:
        print('Error: {}'.format(e))


def automaton_language(text: str, parser: Callable, generator: Callable, length: int) -> str:
    try:
        A = parser(text)
        words = generator(A, length)
        return print_words(words)
    except Exception as e:
        print('Error: {}'.format(e))


def dfa_language(text: str, length: int) -> str:
    return automaton_language(text, parse_dfa, dfa_words_up_to_n, length)


def nfa_language(text: str, length: int) -> str:
    return automaton_language(text, parse_nfa, nfa_words_up_to_n, length)


def pda_language(text: str, length: int) -> str:
    return automaton_language(text, parse_pda, pda_words_up_to_n, length)


def tm_language(text: str, length: int) -> str:
    return automaton_language(text, parse_tm, tm_words_up_to_n, length)


def cfg_language(text: str, length: int) -> str:
    return automaton_language(text, parse_simple_cfg, cfg_words_up_to_n, length)


def regexp_language(text: str, length: int) -> str:
    return automaton_language(text, parse_simple_regexp, regexp_words_up_to_n, length)


# EV, 5jun2020 added check_number_of_nfa_states
def check_number_of_nfa_states(nfa: str, count: int) -> None:
    try:
        N = parse_nfa(nfa)
        if len(N.Q) == count:
            print('OK')
    except Exception as e:
        print('Error: {}'.format(e))


def nfa_accepts(nfa: str, word: str) -> bool:
    try:
        N = parse_nfa(nfa)
        return nfa_accepts_word(N, word)
    except Exception as e:
        print('Error: {}'.format(e))


def regexp_accepts(regex: str, word: str) -> bool:
    try:
        r = parse_simple_regexp(regex)
        return regexp_accepts_word(r, word)
    except Exception as e:
        print('Error: {}'.format(e))


def check_automaton_accepts_rejects(A: Union[DFA, NFA, PDA, TM, CFG, Regexp], accepted: str, rejected: str) -> None:
    def accepts(A: Union[DFA, NFA, PDA, TM, CFG, Regexp], word: str) -> bool:
        if isinstance(A, DFA):
            return dfa_accepts_word(A, word)
        elif isinstance(A, NFA):
            return nfa_accepts_word(A, word)
        elif isinstance(A, PDA):
            return pda_accepts_word(A, word)
        elif isinstance(A, TM):
            return tm_accepts_word(A, word)
        elif isinstance(A, CFG):
            return cfg_accepts_word(A, word)
        elif isinstance(A, Regexp):
            return regexp_accepts_word(A, word)

    accepted_words = parse_word_list(accepted)
    rejected_words = parse_word_list(rejected)

    for word in accepted_words:
        if not accepts(A, word):
            word = word if word else 'ε'
            print("Error: word '{}' should be accepted".format(word))
            return

    for word in rejected_words:
        if accepts(A, word):
            word = word if word else 'ε'
            print("Error: word '{}' should not be accepted".format(word))
            return

    print('OK')


def check_cfg_accepts_rejects(cfg: str, accepted_words: str, rejected_words: str) -> None:
    try:
        G = parse_simple_cfg(cfg)
        check_automaton_accepts_rejects(G, accepted_words, rejected_words)
    except RuntimeError as e:
        print('Error: {}'.format(e))


def check_dfa_accepts_rejects(dfa: str, accepted_words: str, rejected_words: str) -> None:
    D = parse_dfa(dfa)
    check_automaton_accepts_rejects(D, accepted_words, rejected_words)


def check_dfa2regexp(dfa: str, regexp: str, length: int = 8) -> None:
    try:
        D = parse_dfa(dfa)
        R = parse_simple_regexp(regexp)
        feedback = check_equal_languages(R, D, length)
        print_feedback(feedback)
    except Exception as e:
        print('Error: {}'.format(e))



#  (C) Copyright Wieger Wesselink 2020. Distributed under the GPL-3.0-or-later
#  Software License, (See accompanying file LICENSE or copy at
#  https://www.gnu.org/licenses/gpl-3.0.txt)

from typing import List, Set, Tuple, Optional
from collections import defaultdict
import itertools
import io

from gambatools.automaton import Automaton
from gambatools.automaton_algorithms import default_state_label_regex, AutomatonParser, AutomatonBuilder, tm_keywords

from gambatools.dfa import State, Symbol
from gambatools.tm import Direction, TM


def print_tm_state(q, tape, head):
    print(''.join(tape), q)
    print(' ' * head + '^')


def tm_do_transition(T: TM, p: State, tape: List[Symbol], head: int) -> Tuple[State, int]:
    """Does a transition of a Turing Machine, and returns the new state and tape position. The tape is modified."""
    if p in [T.q_reject, T.q_accept]:
        raise RuntimeError('tm_do_transition: can not do a transition in the accepting or rejecting state')

    a = tape[head]
    if (p, a) in T.delta:
        q, b, d = T.delta[p, a]
    else:
        q, b, d = T.q_reject, a, Direction('R')
    tape[head] = b
    head1 = max(head - 1, 0) if d == 'L' else head + 1

    if head1 == len(tape):
        tape.append(T.blank)

    return q, head1


def tm_accepts_word(T: TM, word: str, max_steps: int = 1000) -> Optional[bool]:
    q0 = T.q0
    q_accept = T.q_accept
    q_reject = T.q_reject
    q = q0

    tape = [Symbol(w_i) for w_i in word]
    head = 0
    if head == len(tape):
        tape.append(T.blank)

    for _ in range(max_steps):
        q, head = tm_do_transition(T, q, tape, head)
        if q == q_accept:
            return True
        if q == q_reject:
            return False
    return None


def tm_simulate_word(T: TM, word: str, max_steps: int = 1000) -> List[Tuple[State, List[str], int]]:
    q0 = T.q0
    q_accept = T.q_accept
    q_reject = T.q_reject
    q = q0

    result = []

    tape = [Symbol(w_i) for w_i in word]
    head = 0
    if head == len(tape):
        tape.append(T.blank)
    result.append((q, tape[:], head))

    for _ in range(max_steps):
        q, head = tm_do_transition(T, q, tape, head)
        result.append((q, tape[:], head))
        if q == q_accept:
            break
        if q == q_reject:
            break
    return result


def tm_words_up_to_n(T: TM, n: int, max_steps: int = 1000) -> Set[str]:
    Sigma = T.Sigma
    result = set([])
    for i in range(n + 1):
        for w in itertools.product(Sigma, repeat = i):
            word = ''.join(w)
            if tm_accepts_word(T, word, max_steps):
                result.add(word)
    return result


def print_tm(P: TM) -> str:
    Q = P.Q
    Sigma = P.Sigma
    Gamma = P.Gamma
    delta = P.delta
    q0 = P.q0
    q_accept = P.q_accept
    q_reject = P.q_reject
    blank = P.blank

    out = io.StringIO()
    out.write('states {}\n'.format(' '.join(sorted(Q))))
    out.write('initial {}\n'.format(q0))
    out.write('accept {}\n'.format(q_accept))
    out.write('reject {}\n'.format(q_reject))
    out.write('input_symbols {}\n'.format(' '.join(sorted(Sigma))))
    out.write('tape_symbols {}\n'.format(' '.join(sorted(Gamma))))
    out.write('blank {}\n'.format(blank))
    transitions = defaultdict(lambda: [])
    for (p, a), (q, b, d) in delta.items():
        transitions['{} {}'.format(p, q)].append('{}{},{}'.format(a, b, d))
    for pq in sorted(transitions.keys()):
        out.write('{} {}\n'.format(pq, ' '.join(transitions[pq])))
    result = out.getvalue()
    out.close()
    return result


def tm_transition_transition_regex() -> str:
    s = r'[\w\d~!@#$%^&*□]'
    return r'{0}{0},[LR]'.format(s)


def automaton_to_tm(A: Automaton, transition_regex = tm_transition_transition_regex(), state_regex = default_state_label_regex()) -> TM:
    return TMBuilder(A, state_regex=state_regex, transition_regex=transition_regex).build()


def parse_tm(text: str, transition_regex = tm_transition_transition_regex(), state_regex = default_state_label_regex()) -> TM:
    A = AutomatonParser(state_regex=state_regex, transition_regex=transition_regex, keywords = tm_keywords()).parse(text)
    return automaton_to_tm(A, state_regex=state_regex, transition_regex=transition_regex)


class TMBuilder(AutomatonBuilder):
    def __init__(self, A: Automaton, state_regex=r'\w+', transition_regex=r'\w+'):
        super().__init__(A, state_regex, transition_regex)

    def used_tape_symbols(self) -> Set[str]:
        A = self.A
        result = set([])
        for (_, label, _) in A.transitions:
            result.add(label[0])
            result.add(label[1])
        return result

    def build(self) -> TM:
        A = self.A
        q_accept = self.get_state('accept', self._fresh_state(A.states, 'accept'))
        q_reject = self.get_state('reject', self._fresh_state(A.states, 'reject'))
        A.states = A.states if len(A.states) > 0 else A.used_states() | {q_accept, q_reject}

        self._check_states_are_declared()
        self._check_state_labels()
        self._check_one_initial_state()
        blank = self.parse_symbol('blank', '□', '_')
        tape_symbols = self.get_symbol_set('tape_symbols', self.used_tape_symbols())
        input_symbols = self.get_symbol_set('input_symbols')
        if not input_symbols:
            input_symbols = tape_symbols - {blank}

        Q = set(State(s) for s in A.states)
        Sigma = set(Symbol(s) for s in input_symbols)
        Gamma = set(Symbol(s) for s in tape_symbols)
        delta = defaultdict(set)
        q0 = State(next(iter(A.initial_states)))
        blank = Symbol(blank)
        Gamma.add(blank)
        q_accept = State(q_accept)
        q_reject = State(q_reject)

        for (p, label, q) in A.transitions:
            p = State(p)
            q = State(q)
            a, b, _, d = label
            a = Symbol(a)
            b = Symbol(b)
            d = Direction(d)
            delta[p, a] = (q, b, d)
        return TM(Q, Sigma, Gamma, delta, q0, q_accept, q_reject, blank)



#  (C) Copyright Wieger Wesselink 2020. Distributed under the GPL-3.0-or-later
#  Software License, (See accompanying file LICENSE or copy at
#  https://www.gnu.org/licenses/gpl-3.0.txt)

from typing import Set, Mapping, Tuple, Union

from gambatools.dfa import State, Symbol, print_alphabet, print_state_set, print_delta


class NFA(object):
    def __init__(self, Q: Set[State], Sigma: Set[Symbol], delta: Mapping[Tuple[State, Symbol], Set[State]], q0: State, F: Set[State], epsilon: Symbol = Symbol(''), check_validity: bool = True):
        self.Q = Q
        self.Sigma = Sigma
        self.delta = delta
        self.q0 = q0
        self.F = F
        self.epsilon = epsilon

        if check_validity:
            self._check_validity()

    def _check_validity(self):
        Q = self.Q
        Sigma = self.Sigma
        delta = self.delta
        q0 = self.q0
        F = self.F
        epsilon = self.epsilon
        assert q0 in Q
        assert F <= Q
        assert epsilon not in Sigma
        for (q, a) in delta:
            Q1 = delta[q, a]
            assert q in Q
            assert a in Sigma | {epsilon}
            assert Q1 <= Q

    def E(self, q: Union[State, Set[State]]):
        from gambatools.nfa_algorithms import epsilon_closure
        return epsilon_closure(self, q)

    def __str__(self):
        return _print_nfa(self)


def _print_nfa(N: NFA) -> str:
    return 'Q = {}\nSigma = {}\n{}\nq0 = {}\nF = {}\nepsilon = {}'.format(print_state_set(N.Q), print_alphabet(N.Sigma), print_delta(N.delta), N.q0, print_state_set(N.F), N.epsilon)



#  (C) Copyright Wieger Wesselink 2020. Distributed under the GPL-3.0-or-later
#  Software License, (See accompanying file LICENSE or copy at
#  https://www.gnu.org/licenses/gpl-3.0.txt)

#!/usr/bin/env python3

from gambatools.CFGLexer import *
from gambatools.CFGVisitor import *
from gambatools.CFGParser import *
from gambatools.cfg import *
from gambatools.list_utility import remove_none


class CFGVisitor(CFGVisitor):

    def __init__(self):
        super(CFGVisitor, self).__init__()

    def visitContext_free_grammar(self, ctx: CFGParser.Context_free_grammarContext):
        return self.visit(ctx.rule_list())

    def visitRule_(self, ctx: CFGParser.Rule_Context):
        variable = self.visit(ctx.variable())
        alternatives = self.visit(ctx.alternative_list())
        return (variable, alternatives)

    def visitRule_list(self, ctx: CFGParser.Rule_listContext):
        return remove_none([self.visit(child) for child in ctx.children])

    def visitSymbol(self, ctx: CFGParser.SymbolContext):
        text = ctx.getText()
        return text

    def visitSymbol_list(self, ctx: CFGParser.Symbol_listContext):
        return remove_none([self.visit(child) for child in ctx.children])

    def visitVariable(self, ctx: CFGParser.VariableContext):
        text = ctx.getText()
        return Variable(text)

    def visitAlternative(self, ctx: CFGParser.AlternativeContext):
        return self.visit(ctx.symbol_list())

    def visitAlternative_list(self, ctx: CFGParser.Alternative_listContext):
        return remove_none([self.visit(child) for child in ctx.children])


def parse_cfg(text):
    lexer = CFGLexer(InputStream(text))
    stream = CommonTokenStream(lexer)
    parser = CFGParser(stream)
    grammar = parser.context_free_grammar()
    visitor = CFGVisitor()
    rules = visitor.visit(grammar)

    V = set([rule[0] for rule in rules])
    Sigma = set([])
    R = []
    S = rules[0][0]

    def parse_symbol(text: str) -> Union[Variable, Terminal]:
        if Variable(text) in V:
            return Variable(text)
        else:
            t = Terminal(text)
            if text != '1':
                Sigma.add(t)
            return t

    def parse_alternative(symbols: List[str]) -> Alternative:
        symbols = [parse_symbol(s) for s in symbols]
        if Terminal('1') in symbols:
            if len(symbols) != 1:
                raise RuntimeError('Symbol 1 may only appear in isolation')
            return Alternative([])
        return Alternative(symbols)

    for variable, alternatives in rules:
        for symbols in alternatives:
            rule = Rule(variable, parse_alternative(symbols))
            R.append(rule)
    return CFG(V, Sigma, R, S)



#  (C) Copyright Wieger Wesselink 2020. Distributed under the GPL-3.0-or-later
#  Software License, (See accompanying file LICENSE or copy at
#  https://www.gnu.org/licenses/gpl-3.0.txt)

from typing import Union, Tuple


class Regexp(object):
    pass


class Zero(Regexp):
    def __str__(self):
        return '0'


class One(Regexp):
    def __str__(self):
        return '1'


class Symbol(Regexp):
    def __init__(self, symbol: str):
        self.symbol = symbol

    def __str__(self):
        return self.symbol


class Iteration(Regexp):
    def __init__(self, operand: Regexp):
        self.operand = operand

    def __str__(self):
        return print_unary_right_operation(self, '*')


class Sum(Regexp):
    def __init__(self, left: Regexp, right: Regexp):
        self.left = left
        self.right = right

    def __str__(self):
        return print_binary_operation(self, ' + ')


class Concat(Regexp):
    def __init__(self, left: Regexp, right: Regexp):
        self.left = left
        self.right = right

    def __str__(self):
        return print_binary_operation(self, ' . ')


def precedence(r: Regexp) -> int:
    if isinstance(r, (Zero, One, Symbol)):
        return 10
    elif isinstance(r, Iteration):
        return 9
    elif isinstance(r, Concat):
        return 8
    elif isinstance(r, Sum):
        return 7
    else:
        raise RuntimeError('precedence: unknown expression {}: {}'.format(r, r.__class__))


def is_left_associative(r: Union[Concat, Sum]):
    return True


def is_right_associative(r: Union[Concat, Sum]):
    return True


def print_expression(x: Regexp, needs_parentheses: bool) -> str:
    if needs_parentheses:
        return '({})'.format(x)
    return '{}'.format(x)


def print_unary_operand(x: Iteration, operand: Regexp) -> str:
    return print_expression(operand, precedence(operand) < precedence(x))


def print_unary_left_operation(x: Iteration, op: str) -> str:
    return '{}{}'.format(op, print_unary_operand(x, x.operand))


def print_unary_right_operation(x: Iteration, op: str) -> str:
    return '{}{}'.format(print_unary_operand(x, x.operand), op)


def print_binary_operation(x: Union[Concat, Sum], op: str) -> str:
    x1 = x.left
    x2 = x.right
    p = precedence(x)
    p1 = precedence(x1)
    p2 = precedence(x2)
    return '{}{}{}'.format(
        print_expression(x1, (p1 < p) or (p1 == p and not is_left_associative(x))),
        op,
        print_expression(x2, (p2 < p) or (p2 == p and not is_right_associative(x)))
    )


def print_regexp(x: Regexp) -> str:
    if isinstance(x, Zero):
        return '0'
    elif isinstance(x, One):
        return '1'
    elif isinstance(x, Symbol):
        return x.symbol
    elif isinstance(x, Iteration):
        return '({})*'.format(print_regexp(x.operand))
    elif isinstance(x, Sum):
        return '({} + {})'.format(print_regexp(x.left), print_regexp(x.right))
    elif isinstance(x, Concat):
        return '({} . {})'.format(print_regexp(x.left), print_regexp(x.right))


def print_regexp_simple(x: Regexp) -> str:
    def needs_parentheses_unary(x: Iteration) -> bool:
        return precedence(x.operand) < precedence(x)

    def needs_parentheses_binary(x: Union[Concat, Sum]) -> Tuple[bool, bool]:
        x1 = x.left
        x2 = x.right
        p = precedence(x)
        p1 = precedence(x1)
        p2 = precedence(x2)
        return (p1 < p) or (p1 == p and not is_left_associative(x)), (p2 < p) or (
                    p2 == p and not is_right_associative(x))

    if isinstance(x, Zero):
        return '0'
    elif isinstance(x, One):
        return '1'
    elif isinstance(x, Symbol):
        return x.symbol
    elif isinstance(x, Iteration):
        if needs_parentheses_unary(x):
            return '({})*'.format(print_regexp_simple(x.operand))
        else:
            return '{}*'.format(print_regexp_simple(x.operand))
    elif isinstance(x, Sum):
        n1, n2 = needs_parentheses_binary(x)
        x1 = print_regexp_simple(x.left)
        x2 = print_regexp_simple(x.right)
        if n1:
            x1 = '({})'.format(x1)
        if n2:
            x2 = '({})'.format(x2)
        return '{}+{}'.format(x1, x2)
    elif isinstance(x, Concat):
        n1, n2 = needs_parentheses_binary(x)
        x1 = print_regexp_simple(x.left)
        x2 = print_regexp_simple(x.right)
        if n1:
            x1 = '({})'.format(x1)
        if n2:
            x2 = '({})'.format(x2)
        return '{}{}'.format(x1, x2)



#  (C) Copyright Wieger Wesselink 2020. Distributed under the GPL-3.0-or-later
#  Software License, (See accompanying file LICENSE or copy at
#  https://www.gnu.org/licenses/gpl-3.0.txt)

from typing import NewType, Set, Mapping, Tuple

from gambatools.dfa import State, Symbol, print_alphabet, print_state_set

Direction = NewType('Direction', str)


class TM(object):
    def __init__(self,
                 Q: Set[State],
                 Sigma: Set[Symbol],
                 Gamma: Set[Symbol],
                 delta: Mapping[Tuple[State, Symbol], Tuple[State, Symbol, Direction]],
                 q0: State,
                 q_accept: State,
                 q_reject: State,
                 blank: Symbol = Symbol('_'),
                 check_validity: bool = True):
        self.Q = Q
        self.Sigma = Sigma
        self.Gamma = Gamma
        self.delta = delta
        self.q0 = q0
        self.q_accept = q_accept
        self.q_reject = q_reject
        self.blank = blank

        if check_validity:
            self._check_validity()

    def _check_validity(self):
        Q = self.Q
        Sigma = self.Sigma
        Gamma = self.Gamma
        delta = self.delta
        q0 = self.q0
        q_accept = self.q_accept
        q_reject = self.q_reject
        blank = self.blank
        assert q0 in Q
        assert q_accept in Q
        assert q_reject in Q
        assert q_reject != q_accept
        assert blank not in Sigma
        assert blank in Gamma
        assert Sigma <= Gamma
        for (p, a), (q, b, d) in delta.items():
            assert p in Q
            assert a in Gamma
            assert q in Q
            assert b in Gamma
            assert d in ['L', 'R']
        # total, witness = self.is_total()
        # if not total:
        #     q, a = witness
        #     print('The TM is not total: transition ({},{}) is missing'.format(q, a))

    def is_total(self):
        Q = self.Q
        Gamma = self.Gamma
        delta = self.delta
        blank = self.blank
        q_accept = self.q_accept
        q_reject = self.q_reject
        for q in Q - {q_accept, q_reject}:
            for a in Gamma - {blank}:
                if (q, a) not in delta:
                    return False, (q, a)
        return True, None

    def __str__(self):
        return print_tm(self)


def print_tm_delta(delta: Mapping[Tuple[State, Symbol], Tuple[State, Symbol, Direction]]):
    if len(delta) == 0:
        return 'delta = {:}'
    result = []
    for (p, a), (q, b, d) in delta.items():
        result.append('delta({}, {}) = ({}, {}, {})'.format(p, a, q, b, d))
    return '\n'.join(sorted(result))


def print_tm(T: TM) -> str:
    return 'Q = {}\nSigma = {}\nGamma = {}\n{}\nq0 = {}\nq_accept = {}\nq_reject = {}'.format(print_state_set(T.Q), print_alphabet(T.Sigma), print_alphabet(T.Gamma), print_tm_delta(T.delta), T.q0, T.q_accept, T.q_reject)



#  (C) Copyright Wieger Wesselink 2020. Distributed under the GPL-3.0-or-later
#  Software License, (See accompanying file LICENSE or copy at
#  https://www.gnu.org/licenses/gpl-3.0.txt)

from typing import Set
from gambatools.automaton import Automaton
from gambatools.dfa_algorithms import dfa_words_up_to_n, automaton_to_dfa
from gambatools.nfa_algorithms import nfa_words_up_to_n, automaton_to_nfa


def _compare_words(accepted_words: Set[str], expected_words: Set[str]):
    if 'ε' in expected_words:
        expected_words.remove('ε')
        expected_words.add('')

    if accepted_words != expected_words:
        accepted_minus_expected = accepted_words - expected_words
        if len(accepted_minus_expected) > 0:
            word = next(iter(accepted_minus_expected))
            word = word if word else 'ε'
            feedback = "word '{}' should not be accepted".format(word)
        else:
            word = next(iter(expected_words - accepted_words))
            word = word if word else 'ε'
            feedback = "word '{}' is not accepted".format(word)
        return { 'correct': False, 'feedback': feedback }
    return {'correct': True}


def check_dfa_for_given_language(states, transitions, initial_states, final_states, language: str, max_word_length=5):
    try:
        D = automaton_to_dfa(Automaton(states, transitions, initial_states, final_states, {}))
        expected_words = set(language.split())
        accepted_words = dfa_words_up_to_n(D, max_word_length)
        return _compare_words(accepted_words, expected_words)
    except RuntimeError as e:
        return { 'correct': False, 'feedback': str(e) }

def check_nfa_for_given_language(states, transitions, initial_states, final_states, language: str, max_word_length=5):
    try:
        N = automaton_to_nfa(Automaton(states, transitions, initial_states, final_states, {}))
        expected_words = set(language.split())
        accepted_words = nfa_words_up_to_n(N, max_word_length)
        return _compare_words(accepted_words, expected_words)
    except RuntimeError as e:
        return { 'correct': False, 'feedback': str(e) }



#  (C) Copyright Wieger Wesselink 2020. Distributed under the GPL-3.0-or-later
#  Software License, (See accompanying file LICENSE or copy at
#  https://www.gnu.org/licenses/gpl-3.0.txt)

import itertools
import re
from typing import Set

from gambatools.automaton_algorithms import state_set_regex
from gambatools.dfa import DFA, State, print_state_set
from gambatools.dfa_algorithms import parse_dfa
from gambatools.nfa import NFA
from gambatools.nfa_algorithms import parse_nfa, nfa_to_dfa, epsilon_closure
from gambatools.notebook import show, show_nfa2dfa, print_feedback


def check_nfa_to_dfa_answer(N: NFA, answer: NFA):
    def state(q: Set[State]) -> State:
        return State(print_state_set(q))

    def extract_states(q: State) -> Set[State]:
        label: str = q
        if re.fullmatch(r'{.*}', label):
            label = label[1:-1]
        if label:
            return set(State(s) for s in label.split(','))
        else:
            return set([])

    D = nfa_to_dfa(N)

    feedback = []

    if len(answer.Q) == 0:
        feedback.append('Error: the set of states is empty')

    if N.Sigma != answer.Sigma:
        feedback.append('Error: the alphabet is wrong')

    # check if the states in answer.Q have the right format
    for q in answer.Q:
        if not re.fullmatch(state_set_regex(), q):
            feedback.append("Error: the state label {} has the wrong format".format(q))
            break
        if not extract_states(q) <= N.Q:
            q1 = next(iter(extract_states(q) - N.Q))
            feedback.append("Error: the element {} in {} is not a state of the original NFA".format(q1, q))

    # check the initial state
    if extract_states(answer.q0) != extract_states(D.q0):
        feedback.append('Error: the initial state is incorrect')

    # check the final states
    for q in answer.Q:
        q_states = extract_states(q)
        is_final = not q_states.isdisjoint(N.F)
        if (q in answer.F) != is_final:
            feedback.append('Error: the state {} should {}be marked as final'.format(q, '' if is_final else 'not '))
            break

    # check the transition targets
    for q, a in itertools.product(answer.Q, answer.Sigma):
        q_states = extract_states(q)
        Q1 = answer.delta[q, a]
        if len(Q1) != 1:
            continue
        q1 = next(iter(Q1))
        q1_states = extract_states(q1)

        q1_states_expected = set([])
        for q_ in q_states:
            q1_states_expected |= N.delta[q_, a]
        q1_states_expected = epsilon_closure(N, q1_states_expected)

        if q1_states != q1_states_expected:
            feedback.append('Error: the {}-transition from {} has the wrong target {}'.format(a, print_state_set(q_states), state(q1_states)))
            break

    # check if the states are deterministic and total
    for q, a in itertools.product(answer.Q, answer.Sigma):
            Q1 = answer.delta[q, a]
            if len(Q1) == 0:
                feedback.append('Error: the state {} has no outgoing {}-transition'.format(q, a))
                break
            if len(Q1) > 1:
                feedback.append('Error: the state {} has multiple outgoing {}-transitions'.format(q, a))
                break

    return feedback


def check_nfa2dfa(nfa: str, dfa: str) -> None:
    try:
        N = parse_nfa(nfa)
        answer: NFA = parse_nfa(dfa, state_regex=state_set_regex())  # N.B. We use an NFA, to avoid the strict DFA checking
        feedback = check_nfa_to_dfa_answer(N, answer)
        print_feedback(feedback)
    except Exception as e:
        print('Error: {}'.format(e))



#  (C) Copyright Wieger Wesselink 2020. Distributed under the GPL-3.0-or-later
#  Software License, (See accompanying file LICENSE or copy at
#  https://www.gnu.org/licenses/gpl-3.0.txt)

from typing import Mapping, Set, Tuple

from gambatools.dfa import State, Symbol, print_alphabet, print_state_set
from gambatools.regexp import Regexp


def print_gnfa_delta(delta: Mapping[Tuple[State, State], Regexp]):
    if len(delta) == 0:
        return 'delta = {:}'
    result = []
    for (q1, q2), r in delta.items():
        result.append('delta({}, {}) = {}'.format(q1, q2, r))
    return '\n'.join(sorted(result))


class GNFA(object):
    def __init__(self, Q: Set[State], Sigma: Set[Symbol], delta: Mapping[Tuple[State, State], Regexp], q_start: State, q_accept: State, epsilon: Symbol = Symbol('')):
        self.Q  = Q
        self.Sigma = Sigma
        self.delta = delta
        self.q_start = q_start
        self.q_accept = q_accept
        self.epsilon = epsilon

        self.check_validity()

    def check_validity(self):
        Q = self.Q
        Sigma = self.Sigma
        delta = self.delta
        q_start = self.q_start
        q_accept = self.q_accept
        epsilon = self.epsilon
        # TODO: check if q_start and q_accept should be in Q or not
        # assert q_start not in Q
        # assert q_accept not in Q
        assert epsilon not in Sigma
        for (q1, q2) in delta:
            assert q1 in Q
            assert q2 in Q

    def __str__(self):
        return print_gnfa(self)


def print_gnfa(G: GNFA) -> str:
    return 'Q = {}\nSigma = {}\n{}\nq_start = {}\nq_accept = {}'.format(print_state_set(G.Q), print_alphabet(G.Sigma), print_gnfa_delta(G.delta), G.q_start, G.q_accept)



#  (C) Copyright Wieger Wesselink 2020. Distributed under the GPL-3.0-or-later
#  Software License, (See accompanying file LICENSE or copy at
#  https://www.gnu.org/licenses/gpl-3.0.txt)

from typing import List, Tuple

from gambatools.automaton_algorithms import state_product_regex
from gambatools.dfa_algorithms import parse_dfa, dfa_union, dfa_intersection, dfa_symmetric_difference, \
    dfa_complement, dfa_words_up_to_n
from gambatools.dfa import DFA, State, Symbol
from gambatools.language_algorithms import language_reverse
from gambatools.language_generator import compare_languages, check_equal_languages, generate_language
from gambatools.nfa_algorithms import parse_nfa
from gambatools.notebook import print_feedback, show, show_product


def extract_states(q: State) -> Tuple[State, State]:
    labels = q[1:-1].split(',')
    return State(labels[0]), State(labels[1])


def check_product_automaton(D: DFA, D1: DFA, D2: DFA, answer: DFA) -> List[str]:
    feedback = []

    # check if the states in answer are proper product states
    for q in answer.Q:
        q1, q2 = extract_states(q)
        if q1 not in D1.Q or q2 not in D2.Q:
            feedback.append('The state {} is not a valid product state'.format(q))

    if D.Sigma != answer.Sigma:
        feedback.append('The alphabet of the complement should be equal to the alphabet of the original DFA')

    # check the initial state
    if answer.q0 != D.q0:
        feedback.append('The initial state should be {}'.format(D.q0))

    # check the transitions
    for (q, a), q1 in answer.delta.items():
        if (q, a) in D.delta and q1 != D.delta[q, a]:
            feedback.append('The {}-transition from {} should have target {}'.format(a, q, q1))

    # check the final states
    for q in D.F - answer.F:
        feedback.append('The state {} should be final'.format(q))
    for q in answer.F - D.F:
        feedback.append('The state {} should not be final'.format(q))

    return feedback


def check_dfa_union(dfa: str, dfa1: str, dfa2: str, length: int = 8) -> None:
    from gambatools.language_algorithms import union
    try:
        D1 = parse_dfa(dfa1)
        D2 = parse_dfa(dfa2)
        D = dfa_union(D1, D2)
        answer = parse_dfa(dfa, state_regex=state_product_regex())
        feedback = check_product_automaton(D, D1, D2, answer)
        L1 = dfa_words_up_to_n(D1, length)
        L2 = dfa_words_up_to_n(D2, length)
        L = dfa_words_up_to_n(answer, length)
        feedback = feedback + compare_languages(L, union(L1, L2))
        print_feedback(feedback)
    except Exception as e:
        print('Error: {}'.format(e))


def check_dfa_intersection(dfa: str, dfa1: str, dfa2: str, length: int = 8) -> None:
    from gambatools.language_algorithms import intersection
    try:
        D1 = parse_dfa(dfa1)
        D2 = parse_dfa(dfa2)
        D = dfa_intersection(D1, D2)
        answer = parse_dfa(dfa, state_regex=state_product_regex())
        feedback = check_product_automaton(D, D1, D2, answer)
        L1 = dfa_words_up_to_n(D1, length)
        L2 = dfa_words_up_to_n(D2, length)
        L = dfa_words_up_to_n(answer, length)
        feedback = feedback + compare_languages(L, intersection(L1, L2))
        print_feedback(feedback)
    except Exception as e:
        print('Error: {}'.format(e))


def check_dfa_symmetric_difference(dfa: str, dfa1: str, dfa2: str, length: int = 8) -> None:
    from gambatools.language_algorithms import symmetric_difference
    try:
        D1 = parse_dfa(dfa1)
        D2 = parse_dfa(dfa2)
        D = dfa_symmetric_difference(D1, D2)
        answer = parse_dfa(dfa, state_regex=state_product_regex())
        feedback = check_product_automaton(D, D1, D2, answer)
        L1 = dfa_words_up_to_n(D1, length)
        L2 = dfa_words_up_to_n(D2, length)
        L = dfa_words_up_to_n(answer, length)
        feedback = feedback + compare_languages(L, symmetric_difference(L1, L2))
        print_feedback(feedback)
    except Exception as e:
        print('Error: {}'.format(e))


def check_dfa_complement(dfa: str, dfa1: str, length: int = 8) -> None:
    try:
        D1 = parse_dfa(dfa1)
        D = dfa_complement(D1)
        answer = parse_dfa(dfa)

        feedback = []

        if D.Sigma != answer.Sigma:
            feedback.append('The alphabet of the complement should be equal to the alphabet of the original DFA')

        if D.Q != answer.Q:
            feedback.append('The states of the complement should be equal to the states of the original DFA')

        if D.q0 != answer.q0:
            feedback.append('The initial state of the complement should be equal to the initial state of the original DFA')

        if D.delta != answer.delta:
            feedback.append('The transitions of the complement should be equal to the transitions of the original DFA')

        # check the final states
        for q in D.F - answer.F:
            feedback.append('The state {} should be final'.format(q))
        for q in answer.F - D.F:
            feedback.append('The state {} should not be final'.format(q))

        feedback = []
        print_feedback(feedback)
    except Exception as e:
        print('Error: {}'.format(e))


def check_dfa_reverse(dfa: str, nfa: str, length: int = 8) -> None:
    try:
        D = parse_dfa(dfa)
        answer = parse_nfa(nfa)

        feedback = []

        if D.Sigma != answer.Sigma:
            feedback.append('Error: the alphabet of the complement should be equal to the alphabet of the original DFA')

        if not D.Q <= answer.Q:
            feedback.append('Warning: the states of the original DFA should be reused')

        for (q, a), q1 in D.delta.items():
            if q not in answer.delta[q1, a]:
                feedback.append('Warning: the reversed transition {} --{}-> {} is missing'.format(q1, a, q))
                break

        if answer.q0 in D.Q:
            feedback.append('Warning: a new initial state should be introduced')

        if answer.F != {D.q0}:
            feedback.append('Warning: the initial state of the DFA should be the final state of the NFA')

        L1 = generate_language(answer, length)
        L2 = language_reverse(generate_language(D, length))
        feedback = feedback + compare_languages(L1, L2)

        print_feedback(feedback)

    except Exception as e:
        print('Error: {}'.format(e))





#  (C) Copyright Wieger Wesselink 2020. Distributed under the GPL-3.0-or-later
#  Software License, (See accompanying file LICENSE or copy at
#  https://www.gnu.org/licenses/gpl-3.0.txt)

from typing import NewType, Set, Mapping, Tuple

State = NewType('State', str)
Symbol = NewType('Symbol', str)


def print_alphabet(Sigma: Set[Symbol]) -> str:
    if len(Sigma) == 0:
        return '{}'
    return State('{{ {} }}'.format(', '.join(sorted(Sigma))))


def print_state_set(Q: Set[State]) -> str:
    if len(Q) == 0:
        return '{}'
    return '{{{}}}'.format(','.join(sorted(Q)))


def print_delta(delta):
    if len(delta) == 0:
        return 'delta = {:}'
    result = []
    for (q, a) in delta:
        q1 = delta[q, a]
        q1 = q1 if isinstance(q1, str) else sorted(q1)
        if a == '':
            a = 'epsilon'
        result.append('delta({}, {}) = {}'.format(q, a, q1))
    return '\n'.join(sorted(result))


class DFA(object):
    def __init__(self, Q: Set[State], Sigma: Set[Symbol], delta: Mapping[Tuple[State, Symbol], State], q0: State, F: Set[State], check_validity: bool = True):
        self.Q = Q
        self.Sigma = Sigma
        self.delta = delta
        self.q0 = q0
        self.F = F

        if check_validity:
            self._check_validity()

    def _check_validity(self):
        Q = self.Q
        Sigma = self.Sigma
        delta = self.delta
        q0 = self.q0
        F = self.F
        assert q0 in Q
        assert F <= Q
        for (q, a) in delta:
            q1 = delta[q, a]
            assert q in Q
            assert a in Sigma
            assert q1 in Q
        assert self._is_total()

    def _is_total(self):
        Q = self.Q
        Sigma = self.Sigma
        delta = self.delta
        for q in Q:
            for a in Sigma:
                if (q, a) not in delta:
                    return False
        return True

    def __str__(self):
        return 'Q = {}\nSigma = {}\n{}\nq0 = {}\nF = {}'.format(print_state_set(self.Q), print_alphabet(self.Sigma), print_delta(self.delta), self.q0, print_state_set(self.F))



#  (C) Copyright Wieger Wesselink 2020. Distributed under the GPL-3.0-or-later
#  Software License, (See accompanying file LICENSE or copy at
#  https://www.gnu.org/licenses/gpl-3.0.txt)

from gambatools.draw_sigma import dfa_to_sigma, automaton_to_sigma
from gambatools.automaton_algorithms import parse_automaton
from gambatools.dfa_algorithms import parse_dfa

sigma_container_index = 0


def jupyter_html_settings(sigma_container):
    return {'container': sigma_container, 'style': 'height:800px; background-color:#E1E1E1'}


def show_automaton(text: str) -> str:
    global sigma_container_index
    A = parse_automaton(text)
    sigma_container = 'sigma-container{}'.format(sigma_container_index)
    sigma_container_index = sigma_container_index + 1
    return automaton_to_sigma(A, jupyter_html_settings(sigma_container))


def show_dfa(text: str) -> str:
    global sigma_container_index
    D = parse_dfa(text)
    sigma_container = 'sigma-container{}'.format(sigma_container_index)
    sigma_container_index = sigma_container_index + 1
    return dfa_to_sigma(D, jupyter_html_settings(sigma_container))



#  (C) Copyright Wieger Wesselink 2020. Distributed under the GPL-3.0-or-later
#  Software License, (See accompanying file LICENSE or copy at
#  https://www.gnu.org/licenses/gpl-3.0.txt)

from typing import Set, List, Tuple, Mapping, Any


class Automaton(object):
    def __init__(self,
                 states: Set[str],
                 transitions: List[Tuple[str, str, str]],
                 initial_states: Set[str],
                 final_states: Set[str],
                 items: Mapping[str, List[Any]]
                ):
        self.states = states
        self.transitions = transitions
        self.initial_states = initial_states
        self.final_states = final_states
        self.items = items

    def used_states(self) -> Set[str]:
        result = self.initial_states | self.final_states
        for (p, _, q) in self.transitions:
            result.add(p)
            result.add(q)
        return result



#  (C) Copyright Wieger Wesselink 2020. Distributed under the GPL-3.0-or-later
#  Software License, (See accompanying file LICENSE or copy at
#  https://www.gnu.org/licenses/gpl-3.0.txt)

from typing import Set, MutableMapping, Tuple, Mapping

from gambatools.dfa import State, Symbol, print_alphabet, print_state_set


class PDA(object):
    def __init__(self,
                 Q: Set[State],
                 Sigma: Set[Symbol],
                 Gamma: Set[Symbol],
                 delta: MutableMapping[Tuple[State, Symbol, Symbol], Set[Tuple[State, Symbol]]],
                 q0: State,
                 F: Set[State],
                 epsilon: Symbol = Symbol(''),
                 check_validity: bool = True):
        self.Q = Q
        self.Sigma = Sigma
        self.Gamma = Gamma
        self.delta = delta
        self.q0 = q0
        self.F = F
        self.epsilon = epsilon

        if check_validity:
            self._check_validity()

    def _check_validity(self):
        Q = self.Q
        Sigma = self.Sigma
        Gamma = self.Gamma
        delta = self.delta
        q0 = self.q0
        F = self.F
        epsilon = self.epsilon
        assert q0 in Q
        assert epsilon not in Sigma
        assert epsilon not in Gamma
        assert F <= Q
        for (p, a, u), Q1 in delta.items():
            assert p in Q
            assert a in Sigma | {epsilon}
            assert u in Gamma | {epsilon}
            for (q, v) in Q1:
                assert q in Q
                assert v in Gamma | {epsilon}

    def is_push_pop_transition(self, p: State, a: Symbol, u: Symbol, q: State, v: Symbol) -> bool:
        epsilon = self.epsilon
        return (u == epsilon and v != epsilon) or (u != epsilon and v == epsilon)

    def __str__(self):
        return print_pda(self)


def print_pda_delta(delta: Mapping[Tuple[State, Symbol, Symbol], Set[Tuple[State, Symbol]]]):
    if len(delta) == 0:
        return 'delta = {:}'
    result = []
    for (p, a, u), Q1 in delta.items():
        result.append('delta({}, {}, {}) = {}'.format(p, a, u, Q1))
    return '\n'.join(sorted(result))


def print_pda(P: PDA) -> str:
    return 'Q = {}\nSigma = {}\nGamma = {}\n{}\nq0 = {}\nF = {}'.format(print_state_set(P.Q), print_alphabet(P.Sigma), print_alphabet(P.Gamma), print_pda_delta(P.delta), P.q0, print_state_set(P.F))



#  (C) Copyright Wieger Wesselink 2020. Distributed under the GPL-3.0-or-later
#  Software License, (See accompanying file LICENSE or copy at
#  https://www.gnu.org/licenses/gpl-3.0.txt)

import random

from IPython.core.display import HTML

from gambatools.cfg_algorithms import parse_simple_cfg, cfg_accepts_word
from gambatools.language_algorithms import parse_word_list
from gambatools.printing import print_words


def hide_code():
    this_cell = """$('div.cell.code_cell.rendered.selected')"""

    toggle_text = '.'  # text shown on toggle link
    target_cell = this_cell  # target cell to control with toggle
    js_hide_current = ''  # bit of JS to permanently hide code in current cell (only when toggling next cell)

    js_f_name = 'code_toggle_{}'.format(str(random.randint(1,2**64)))

    html = """
        <script>
            function {f_name}() {{
                {cell_selector}.find('div.input').toggle();
            }}

            {js_hide_current}
        </script>

        <a href="javascript:{f_name}()">{toggle_text}</a>
    """.format(
        f_name=js_f_name,
        cell_selector=target_cell,
        js_hide_current=js_hide_current,
        toggle_text=toggle_text
    )

    return HTML(html)


def check_cfg_accepts(cfg: str, word_list: str):
    try:
        G = parse_simple_cfg(cfg)
        words = parse_word_list(word_list)
        failures = []
        for word in words:
            if not cfg_accepts_word(G, word):
                failures.append(word)
        if failures:
            print('The following words should be accepted: {}'.format(print_words(failures)))
        else:
            print('OK')
    except RuntimeError as e:
        print('Error: {}'.format(e))


def check_cfg_rejects(cfg: str, word_list: str):
    G = parse_simple_cfg(cfg)
    words = parse_word_list(word_list)
    failures = []
    for word in words:
        if cfg_accepts_word(G, word):
            failures.append(word)
    if failures:
        print('The following words should be rejected: {}'.format(print_words(failures)))
    else:
        print('OK')





# Generated from regexp_simple.g4 by ANTLR 4.8
from antlr4 import *
if __name__ is not None and "." in __name__:
    from .regexp_simpleParser import regexp_simpleParser
else:
    from regexp_simpleParser import regexp_simpleParser

# This class defines a complete generic visitor for a parse tree produced by regexp_simpleParser.

class regexp_simpleVisitor(ParseTreeVisitor):

    # Visit a parse tree produced by regexp_simpleParser#OneExpression.
    def visitOneExpression(self, ctx:regexp_simpleParser.OneExpressionContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by regexp_simpleParser#SymbolExpression.
    def visitSymbolExpression(self, ctx:regexp_simpleParser.SymbolExpressionContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by regexp_simpleParser#IterationExpression.
    def visitIterationExpression(self, ctx:regexp_simpleParser.IterationExpressionContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by regexp_simpleParser#ConcatExpression.
    def visitConcatExpression(self, ctx:regexp_simpleParser.ConcatExpressionContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by regexp_simpleParser#SumExpression.
    def visitSumExpression(self, ctx:regexp_simpleParser.SumExpressionContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by regexp_simpleParser#ZeroExpression.
    def visitZeroExpression(self, ctx:regexp_simpleParser.ZeroExpressionContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by regexp_simpleParser#ParensExpression.
    def visitParensExpression(self, ctx:regexp_simpleParser.ParensExpressionContext):
        return self.visitChildren(ctx)



del regexp_simpleParser


#  (C) Copyright Wieger Wesselink 2020. Distributed under the GPL-3.0-or-later
#  Software License, (See accompanying file LICENSE or copy at
#  https://www.gnu.org/licenses/gpl-3.0.txt)

import re
from typing import Set, List, Union

from IPython.display import display, Markdown
from gambatools.cfg import Variable, Terminal, CFG, Rule
from gambatools.cfg_algorithms import parse_simple_cfg, cfg_cyk_matrix
from gambatools.notebook import print_feedback


def check_cyk_matrix(cfg: str, word: str, answer: str) -> None:
    def print_set(S: Set[str]) -> str:
        return '{' + ','.join(list(map(str, sorted(S)))) + '}'

    try:
        feedback = []
        G = parse_simple_cfg(cfg)
        Y = cfg_cyk_matrix(G, word)

        # check the syntax of the items
        lines = re.split('\n', answer.strip())
        for i, line in enumerate(lines):
            words = line.strip().split()
            for w in words:
                if w != '{}' and not re.fullmatch(r'{\w(,\w)*}', w):
                    feedback.append('Error: the entry {} is ill formed'.format(w))
                else:
                    variables = set(Variable(x) for x in re.sub(r'[{},]', '', w))
                    if not variables <= G.V:
                        feedback.append('Error: the entry {} contains unknown variables'.format(w))

        # check the sizes
        for i, line in enumerate(lines):
            words = line.strip().split()
            if len(words) != i + 1:
                feedback.append('Error: line {0} should contain {0} entries'.format(i + 1))

        # check the content; report at most one error
        lines = reversed(lines)
        n = len(word)
        for i, line in enumerate(lines):
            if feedback:
                break
            words = line.strip().split()
            for j, w in enumerate(words):
                expected_variables = Y[j, i+j]
                variables = set(Variable(x) for x in re.sub(r'[{},]', '', w))
                if variables != expected_variables:
                    feedback.append('Error: X[{},{}] has the wrong value {} instead of {}'.format(j+1, i+j+1, w, print_set(expected_variables)))
                    break

        print_feedback(feedback)
        if not feedback:
            accepted = G.S in Y[0, n - 1]
            markdown = 'The start variable {} is contained in $X_{{1n}}$, hence the word is {}accepted.'.format(G.S, '' if accepted else 'not ')
            display(Markdown(markdown))
            # print('The start variable {} is contained in X[1,{}], hence the word is {}accepted.'.format(G.S, n, '' if accepted else 'not '))
    except Exception as e:
        print('Error: {}'.format(e))


def cfg_apply_rule(G: CFG, rule: Rule, element: List[Union[Variable, Terminal]], derivation_type: str) -> List[List[Union[Variable, Terminal]]]:
    from gambatools.algorithms import last_index

    def apply(pos: int) -> List[Union[Variable, Terminal]]:
        return element[:pos] + rule.alternative.symbols + element[pos+1:]

    if rule.variable not in element:
        return []
    elif derivation_type == 'leftmost':
        pos = element.index(rule.variable)
        return [apply(pos)]
    elif derivation_type == 'rightmost':
        pos = last_index(element, rule.variable)
        return [apply(pos)]
    elif derivation_type == 'any':
        return [apply(pos) for pos, e in enumerate(element) if e == rule.variable]
    else:
        raise RuntimeError('unknown derivation type {}'.format(derivation_type))


# returns True if there is a derivation of the given type from elem1 to elem2
def cfg_has_derivation(G: CFG, elem1: List[Union[Variable, Terminal]], elem2: List[Union[Variable, Terminal]], derivation_type: str) -> bool:
    def find_variables() -> Set[Variable]:
        variables = [v for v in elem1 if isinstance(v, Variable)]
        if not variables:
            return set([])
        elif derivation_type == 'any':
            return set(variables)
        elif derivation_type == 'leftmost':
            return {variables[0]}
        elif derivation_type == 'rightmost':
            return {variables[-1]}
        else:
            raise RuntimeError('unknown derivation type {}'.format(derivation_type))

    R = G.R
    variables = find_variables()

    for rule in R:
        if rule.variable not in variables:
            continue
        if elem2 in cfg_apply_rule(G, rule, elem1, derivation_type):
            return True
    return False


def check_cfg_derivation(cfg: str, derivation: str, word: str, derivation_type='leftmost') -> None:
    def parse_character(c: str) -> Union[Variable, Terminal]:
        return Variable(c) if c.isupper() else Terminal(c)

    try:
        feedback = []
        G = parse_simple_cfg(cfg)
        V = G.V
        S = G.S
        Sigma = G.Sigma

        # parse the elements of the derivation
        words = derivation.strip().split('=>')
        words = [w.strip() for w in words]
        elements = [list(map(parse_character, element)) for element in words]

        if not elements:
            feedback.append('Error: the derivation is empty')

        # check if the elements correspond to the grammar
        for element in elements:
            for e in element:
                if (isinstance(e, Variable) and e not in V) or (isinstance(e, Terminal) and e not in Sigma):
                    feedback.append('Error: the element {} is invalid'.format(''.join(list(map(str, element)))))

        # check if the first element is the start symbol
        if elements:
            first = elements[0]
            if len(first) != 1 or first[0] != S:
                feedback.append('Error: the first element of the derivation must be the start variable of the grammar')

        # check if the steps are valid derivation steps
        msg = '' if derivation_type == 'any' else derivation_type + ' '
        for i in range(len(elements) - 1):
            if not cfg_has_derivation(G, elements[i], elements[i+1], derivation_type):
                feedback.append('Error: there is no {}derivation from {} to {}'.format(msg, words[i], words[i+1]))

        # check if the last element is equal to word
        if elements:
            last = elements[-1]
            if last != list(map(Terminal, word)):
                feedback.append('Error: the last element of the derivation must be equal to {}'.format(word))

        print_feedback(feedback)

    except Exception as e:
        print('Error: {}'.format(e))



# Generated from regexp.g4 by ANTLR 4.8
from antlr4 import *
if __name__ is not None and "." in __name__:
    from .regexpParser import regexpParser
else:
    from regexpParser import regexpParser

# This class defines a complete generic visitor for a parse tree produced by regexpParser.

class regexpVisitor(ParseTreeVisitor):

    # Visit a parse tree produced by regexpParser#OneExpression.
    def visitOneExpression(self, ctx:regexpParser.OneExpressionContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by regexpParser#SymbolExpression.
    def visitSymbolExpression(self, ctx:regexpParser.SymbolExpressionContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by regexpParser#IterationExpression.
    def visitIterationExpression(self, ctx:regexpParser.IterationExpressionContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by regexpParser#ConcatExpression.
    def visitConcatExpression(self, ctx:regexpParser.ConcatExpressionContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by regexpParser#SumExpression.
    def visitSumExpression(self, ctx:regexpParser.SumExpressionContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by regexpParser#ZeroExpression.
    def visitZeroExpression(self, ctx:regexpParser.ZeroExpressionContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by regexpParser#ParensExpression.
    def visitParensExpression(self, ctx:regexpParser.ParensExpressionContext):
        return self.visitChildren(ctx)



del regexpParser


#  (C) Copyright Wieger Wesselink 2020. Distributed under the GPL-3.0-or-later
#  Software License, (See accompanying file LICENSE or copy at
#  https://www.gnu.org/licenses/gpl-3.0.txt)

class IdentifierGenerator(object):
    def __init__(self, index: int = 0):
        self.index = index

    def generate(self, hint):
        id = '{}{}'.format(hint, self.index)
        self.index = self.index + 1
        return id


#  (C) Copyright Wieger Wesselink 2020. Distributed under the GPL-3.0-or-later
#  Software License, (See accompanying file LICENSE or copy at
#  https://www.gnu.org/licenses/gpl-3.0.txt)

from typing import List, Set, Iterable, Optional, Tuple
from collections import defaultdict
import copy
import itertools
import io

from gambatools.automaton import Automaton
from gambatools.automaton_algorithms import default_state_label_regex, AutomatonParser, AutomatonBuilder, pda_keywords

from gambatools.dfa import State, Symbol
from gambatools.dfa_algorithms import fresh_state
from gambatools.pda import PDA
from gambatools.cfg import CFG, Variable, Terminal, Rule, Alternative


def pda_is_push_pop(P: PDA) -> bool:
    """Returns true if the following property holds:
       Each transition either pushes a symbol onto the stack (a push move) or pops
       one off the stack (a pop move), but it does not do both at the same time.
    :param P:
    :return:
    """
    delta = P.delta
    return all([P.is_push_pop_transition(p, a, u, q, v)  for (p, a, u), Q1 in delta.items() for (q, v) in Q1])


# TODO: implement a more robust solution
def fresh_symbol(Sigma: Set[str], symbols: Iterable[str]) -> Symbol:
    for s in symbols:
        symbol = Symbol(s)
        if symbol not in Sigma:
            return symbol
    raise RuntimeError('Could not find a fresh symbol in {}'.format(symbols))


def pda_to_one_accepting_state_in_place(P: PDA) -> None:
    """Modifies P such that it has exactly one accepting state"""

    Q = P.Q
    delta = P.delta
    F = P.F
    epsilon = P.epsilon

    if len(F) == 1:
        return

    q_accept = fresh_state(Q, 'q_accept')
    Q.add(q_accept)

    for q in F:
        delta[q, epsilon, epsilon].add((q_accept, epsilon))

    F.clear()
    F.add(q_accept)


def pda_to_accept_on_empty_stack_in_place(P: PDA) -> None:
    """Modifies P such that it only accepts on empty stack"""

    Q = P.Q
    q0 = P.q0
    delta = P.delta
    Gamma = P.Gamma
    F = P.F
    epsilon = P.epsilon

    stack_bottom = fresh_symbol(Gamma, '$@#*&!?')
    Gamma.add(stack_bottom)

    # define a new initial state
    q_initial = fresh_state(Q, 'q_initial')
    Q.add(q_initial)
    delta[q_initial, epsilon, epsilon].add((q0, stack_bottom))
    P.q0 = q_initial

    # define a new accepting state
    q_accept = fresh_state(Q, 'q_accept')
    Q.add(q_accept)
    for q in F:
        delta[q, epsilon, stack_bottom].add((q_accept, epsilon))
    F.clear()
    F.add(q_accept)


def pda_to_accept_on_empty_stack(P: PDA) -> PDA:
    P = copy.deepcopy(P)
    pda_to_accept_on_empty_stack_in_place(P)
    return P


def pda_to_push_pop_in_place(P: PDA) -> None:
    """Brings the PDA P in push/pop format"""

    Q = P.Q
    Gamma = P.Gamma
    delta = P.delta
    F = P.F
    epsilon = P.epsilon

    pda_to_one_accepting_state_in_place(P)

    # add intermediate states to enforce push/pop transitions
    dummy = Symbol('∅')
    assert dummy not in Gamma # TODO: implement a robust solution
    Gamma.add(dummy)
    delta1 = defaultdict(lambda: set([]))
    for (p, a, u), Q1 in delta.items():
        for (q, v) in Q1:
            if P.is_push_pop_transition(p, a, u, q, v):
                delta1[p, a, u].add((q, v))
            elif u == epsilon and v == epsilon:
                q_mid = fresh_state(Q, 'M')
                Q.add(q_mid)
                delta1[p, a, epsilon].add((q_mid, dummy))
                delta1[q_mid, epsilon, dummy].add((q, epsilon))
            elif u != epsilon and v != epsilon:
                q_mid = fresh_state(Q, 'M')
                Q.add(q_mid)
                delta1[p, a, u].add((q_mid, epsilon))
                delta1[q_mid, epsilon, epsilon].add((q, v))
    delta.clear()
    delta.update(delta1)


def pda_to_push_pop(P: PDA) -> PDA:
    """Brings the PDA P in push/pop format"""
    P = copy.deepcopy(P)
    pda_to_push_pop_in_place(P)
    return P


def pda_to_cfg(P: PDA, accepts_on_empty_stack: bool = False) -> CFG:
    """Implements the algorithm in Sipser 3rd ed. page 122.
    """

    P = copy.deepcopy(P)

    if len(P.F) != 1:
        pda_to_one_accepting_state_in_place(P)

    if not pda_is_push_pop(P):
        pda_to_push_pop_in_place(P)

    if not accepts_on_empty_stack:
        pda_to_accept_on_empty_stack_in_place(P)

    def variable(p: State, q: State) -> Variable:
        return Variable("{}'{}".format(p, q))

    Q = P.Q
    Sigma = P.Sigma
    Gamma = P.Gamma
    delta = P.delta
    q0 = P.q0
    F = P.F
    epsilon = Terminal(P.epsilon)
    q_accept = next(iter(F))

    def make_rule(A, a):
        return Rule(A, Alternative([value for value in a if value != epsilon]))

    V: Set[Variable] = {variable(p, q) for (p, q) in itertools.product(Q, repeat=2)}
    Sigma1: Set[Terminal] = set([Terminal(a) for a in Sigma])
    R: List[Rule] = []
    S: Variable = variable(q0, q_accept)

    T_pop = defaultdict(lambda: [])
    T_push = defaultdict(lambda: [])
    for (p, a, u), Q1 in delta.items():
        for (q, v) in Q1:
            if u == epsilon:
                T_push[v].append((p, a, u, q, v))
            else:
                T_pop[u].append((p, a, u, q, v))

    for u in Gamma:
        for (p, a, _, r, _), (s, b, _, q, _) in itertools.product(T_push[u], T_pop[u]):
            Apq = variable(p, q)
            Ars = variable(r, s)
            a_ = Terminal(a)
            b_ = Terminal(b)
            R.append(make_rule(Apq, [a_, Ars, b_]))

    for (p, q, r) in itertools.product(Q, repeat=3):
        Apq = variable(p, q)
        Apr = variable(p, r)
        Arq = variable(r, q)
        R.append(make_rule(Apq, [Apr, Arq]))

    for p in Q:
        App = variable(p, p)
        R.append(make_rule(App, []))

    return CFG(V, Sigma1, R, S, epsilon)


class PDAState(object):
    def __init__(self, q: State, stack: List[Symbol]):
        self.q = q
        self.stack = stack[:]

    def __str__(self):
        return '({}, [{}])'.format(self.q, ''.join(self.stack))

    def __eq__(self, other):
        return self.q == other.q and self.stack == other.stack

    def __hash__(self):
        return hash(str(self)) # TODO: improve the efficiency

    def __lt__(self, other):
        return str(self) < str(other)  # TODO: improve the efficiency


def pda_can_pop_push(P: PDA, stack: List[Symbol], u: Symbol, v: Symbol) -> bool:
    """Returns true if u can be popped from and then v pushed on stack"""
    return u == P.epsilon or (stack and stack[-1] == u)


def pda_pop_push(P: PDA, stack: List[Symbol], u: Symbol, v: Symbol) -> List[Symbol]:
    """Tries to pop u from and push v on stack. Returns a copy of the resulting stack"""
    epsilon = P.epsilon
    if u == epsilon:
        if v == epsilon:
            return stack[:]
        else:
            return stack + [v]
    else:
        if not stack or stack[-1] != u:
            raise RuntimeError('pda_pop_push({}, {}, {}) is not possible!'.format(stack, u, v))
        if v == epsilon:
            return stack[:-1]
        else:
            return stack[:-1] + [v]


def pda_epsilon_closure(P: PDA, R: Iterable[PDAState]) -> Set[PDAState]:
    """Returns all states reachable from an element in R via epsilon steps."""
    delta = P.delta
    epsilon = P.epsilon

    result: Set[PDAState] = set([r for r in R])
    todo: Set[PDAState] = set([r for r in R])
    while len(todo) > 0:
        src = todo.pop()
        for (p, a, u), Q1 in delta.items():
            if p != src.q or a != epsilon:
                continue
            for (q, v) in Q1:
                if pda_can_pop_push(P, src.stack, u, v):
                    stack1 = pda_pop_push(P, src.stack, u, v)
                    target = PDAState(q, stack1)
                    if target not in result:
                        todo.add(target)
                        result.add(target)
    return result


def pda_do_transition(P: PDA, a: Symbol, R: Iterable[PDAState]) -> Set[PDAState]:
    """Returns all PDA states reachable from an element in R via an a-transition."""
    delta = P.delta

    result: Set[PDAState] = set([])
    for r in R:
        p, stack = r.q, r.stack
        for (p1, a1, u), Q1 in delta.items():
            if p != p1 or a != a1:
                continue
            for (q, v) in Q1:
                if pda_can_pop_push(P, stack, u, v):
                    stack1 = pda_pop_push(P, stack, u, v)
                    r1 = PDAState(q, stack1)
                    # print('transition {}{} {},{}->{}'.format(p1, q, a1, u, v))
                    # print('p = {}'.format(p))
                    # print('r1 = {}'.format(r1))
                    result.add(r1)
    return result


def pda_find_epsilon_path(P: PDA, R: Set[PDAState], f: PDAState) -> Optional[List[PDAState]]:
    """Returns a path from an element r in R to f"""
    delta = P.delta
    epsilon = P.epsilon

    if f in R:
        return [f]

    backpointers = {}

    def make_path(q: PDAState) -> List[PDAState]:
        path = [q]
        while q not in R:
            q = backpointers[q]
            path.insert(0, q)
        return path

    visited: Set[PDAState] = set([r for r in R])
    todo: Set[PDAState] = set([r for r in R])
    while len(todo) > 0:
        src = todo.pop()
        for (p, a, u), Q1 in delta.items():
            if p != src.q or a != epsilon:
                continue
            for (q, v) in Q1:
                if pda_can_pop_push(P, src.stack, u, v):
                    stack1 = pda_pop_push(P, src.stack, u, v)
                    target = PDAState(q, stack1)
                    backpointers[target] = src
                    if target == f:
                        return make_path(target)
                    if target not in visited:
                        todo.add(target)
                        visited.add(target)
    return None


def pda_find_transition(P: PDA, R: Set[PDAState], a: Symbol, target: PDAState) -> Optional[PDAState]:
    """Returns src such that src --a--> target"""
    delta = P.delta
    for src in R:
        for (p, a1, u), Q1 in delta.items():
            if src.q != p or a != a1:
                continue
            for (q, v) in Q1:
                if q == target.q and pda_can_pop_push(P, src.stack, u, v) and pda_pop_push(P, src.stack, u, v) == target.stack:
                    return src
    return None


def pda_accepts_word(P: PDA, w: str) -> bool:
    F = P.F
    R = {PDAState(P.q0, [])}
    R = pda_epsilon_closure(P, R)
    for a in w:
        R = pda_do_transition(P, Symbol(a), R)
        R = pda_epsilon_closure(P, R)
    return any(r.q in F for r in R)


def pda_simulate_word(P: PDA, w: str) -> Optional[Tuple[State, str, List[Symbol]]]:
    F = P.F
    H = []

    def make_row(r: PDAState, word: str):
        return (r.q, word, r.stack[:])

    R = {PDAState(P.q0, [])}
    H.append(R)
    R = pda_epsilon_closure(P, R)
    H.append(R)

    for a in w:
        R = pda_do_transition(P, Symbol(a), R)
        H.append(R)
        R = pda_epsilon_closure(P, R)
        H.append(R)

    if any(r.q in F for r in R):
        S = H.pop()
        front = next(r for r in S if r.q in F)
        result = [make_row(front, '')]
        word = ''
        for a in reversed(w):
            S = H.pop()
            path = pda_find_epsilon_path(P, S, front)
            result = [make_row(r, word) for r in path[:-1]] + result
            front = path[0]
            S = H.pop()
            front = pda_find_transition(P, S, a, front)
            word = a + word
            result = [make_row(front, word)] + result
        S = H.pop()
        path = pda_find_epsilon_path(P, S, front)
        result = [make_row(r, word) for r in path[:-1]] + result
        return result
    else:
        return None


def pda_words_up_to_n(P: PDA, n: int) -> Set[str]:
    F = P.F
    Sigma = P.Sigma

    result = set([])
    W = defaultdict(lambda: set([]))
    R = {PDAState(P.q0, [])}
    R = pda_epsilon_closure(P, R)
    for r in R:
        W[r] = {''}
        if r.q in F:
            result.add('')

    for i in range(n):
        W1 = defaultdict(lambda: set([]))
        for r, words in W.items():
            for a in Sigma:
                R = pda_do_transition(P, a, {r})
                R = pda_epsilon_closure(P, R)
                words_plus_a = set(word + a for word in words)
                for r1 in R:
                    W1[r1] |= words_plus_a
                    if r1.q in F:
                        result |= words_plus_a
        W = W1
    return result


def print_pda(P: PDA) -> str:
    Q = P.Q
    Sigma = P.Sigma
    Gamma = P.Gamma
    delta = P.delta
    q0 = P.q0
    F = P.F
    epsilon = P.epsilon

    out = io.StringIO()
    out.write('states {}\n'.format(' '.join(sorted(Q))))
    out.write('final {}\n'.format(' '.join(sorted(F))))
    out.write('initial {}\n'.format(q0))
    out.write('input_symbols {}\n'.format(' '.join(sorted(Sigma))))
    out.write('stack_symbols {}\n'.format(' '.join(sorted(Gamma))))
    out.write('epsilon {}\n'.format(epsilon))
    transitions = defaultdict(lambda: [])
    for (p, a, u), Q1 in delta.items():
        for (q, v) in Q1:
            transitions['{} {}'.format(p, q)].append('{},{}{}'.format(a, u, v))
    for pq in sorted(transitions.keys()):
        out.write('{} {}\n'.format(pq, ' '.join(transitions[pq])))
    result = out.getvalue()
    out.close()
    return result


def pda_transition_transition_regex() -> str:
    s = r'[\w\d~!@#$%^&*]'
    return r'\w,{0}{0}'.format(s)


def automaton_to_pda(A: Automaton, transition_regex = pda_transition_transition_regex(), state_regex = default_state_label_regex()) -> PDA:
    return PDABuilder(A, state_regex=state_regex, transition_regex=transition_regex).build()


def parse_pda(text: str, transition_regex = pda_transition_transition_regex(), state_regex = default_state_label_regex()) -> PDA:
    A = AutomatonParser(state_regex=state_regex, transition_regex=transition_regex, keywords = pda_keywords()).parse(text)
    return automaton_to_pda(A, state_regex=state_regex, transition_regex=transition_regex)


class PDABuilder(AutomatonBuilder):
    def __init__(self, A: Automaton, state_regex=r'\w+', transition_regex=r'\w+'):
        super().__init__(A, state_regex, transition_regex)

    def used_input_symbols(self, epsilon: str) -> Set[str]:
        A = self.A
        return set(label[0] for (_, label, _) in A.transitions if label[0] != epsilon)

    def used_stack_symbols(self, epsilon: str) -> Set[str]:
        A = self.A
        result = set([])
        for (_, label, _) in A.transitions:
            result.add(label[2])
            result.add(label[3])
        return result - {epsilon}

    def build(self) -> PDA:
        A = self.A
        A.states = A.states if len(A.states) > 0 else A.used_states()

        self._check_states_are_declared()
        self._check_state_labels()
        self._check_one_initial_state()
        epsilon = self.parse_symbol()
        input_symbols = self.get_symbol_set('input_symbols', self.used_input_symbols(epsilon))
        stack_symbols = self.get_symbol_set('stack_symbols', self.used_stack_symbols(epsilon))
        self._check_symbols(input_symbols)

        Q = set(State(s) for s in A.states)
        Sigma = set(Symbol(s) for s in input_symbols)
        Gamma = set(Symbol(s) for s in stack_symbols)
        delta = defaultdict(set)
        q0 = State(next(iter(A.initial_states)))
        F = set(State(s) for s in A.final_states)
        epsilon = Symbol(epsilon)

        for (p, label, q) in A.transitions:
            p = State(p)
            q = State(q)
            a, _, u, v = label
            a = Symbol(a)
            u = Symbol(u)
            v = Symbol(v)
            delta[p, a, u].add((q, v))
        return PDA(Q, Sigma, Gamma, delta, q0, F, epsilon)



#  (C) Copyright Wieger Wesselink 2020. Distributed under the GPL-3.0-or-later
#  Software License, (See accompanying file LICENSE or copy at
#  https://www.gnu.org/licenses/gpl-3.0.txt)

from typing import Set
import itertools


def intersection(L1: Set[str], L2: Set[str]) -> Set[str]:
    """Returns the intersection of L1 and L2"""
    return L1 & L2


def union(L1: Set[str], L2: Set[str]) -> Set[str]:
    """Returns the union of L1 and L2"""
    return L1 | L2


def symmetric_difference(L1: Set[str], L2: Set[str]) -> Set[str]:
    """Returns the symmetric difference of L1 and L2"""
    return L1 ^ L2


def concatenation(L1: Set[str], L2: Set[str]) -> Set[str]:
    """Returns the concatenation of L1 and L2"""
    return set(w1 + w2 for (w1, w2) in itertools.product(L1, L2))


def language_reverse(L: Set[str]) -> Set[str]:
    """Returns the reverse of L"""
    return set(w[::-1] for w in L)


def words_of_length_n(Sigma: Set[str], n: int) -> Set[str]:
    """Returns all words w in Sigma^* with length n"""
    return set(''.join(x) for x in itertools.product(Sigma, repeat=n))


def words_up_to_n(Sigma: Set[str], n: int) -> Set[str]:
    """Returns all words w in Sigma^* with length <= n"""
    return set().union(*[words_of_length_n(Sigma, i) for i in range(n+1)])


def language_no_prefix(L: Set[str]) -> Set[str]:
    """Returns all words w in L such that no proper prefix of w is in L"""

    def has_prefix_in_L(w: str) -> bool:
        return any(w[i:] in L for i in range(1, len(w)))

    return set(w for w in L if not has_prefix_in_L(w))


def language_no_extend(L: Set[str]) -> Set[str]:
    """Returns all words w in L such that w is not a proper prefix of any word in L"""

    def is_proper_prefix(v: str, w: str) -> bool:
        return w.startswith(v) and w != v

    return set(w for w in L if all(not is_proper_prefix(w, v) for v in L))


# def words_up_to_n(Sigma: Set[str], n: int) -> Set[str]:
#     """Returns all words w in Sigma^* with length at most n"""
#     return set().union(*words_up_to_n(Sigma, k) for k in range(n + 1))


def parse_word_list(word_list: str) -> Set[str]:
    words = [word.strip() for word in word_list.strip().split()]
    words = ['' if word in ['ε', '_'] else word for word in words]
    return set(words)



#  (C) Copyright Wieger Wesselink 2020. Distributed under the GPL-3.0-or-later
#  Software License, (See accompanying file LICENSE or copy at
#  https://www.gnu.org/licenses/gpl-3.0.txt)

import copy
import itertools
import io
from collections import defaultdict
from typing import Set, MutableMapping, Tuple, Mapping, List

from gambatools.automaton import Automaton
from gambatools.automaton_algorithms import default_transition_label_regex, default_state_label_regex, AutomatonParser, AutomatonBuilder
from gambatools.dfa import State, Symbol, print_state_set, DFA
from gambatools.nfa import NFA


def dfa_accepts_word(D: DFA, word: str) -> bool:
    delta = D.delta
    q0 = D.q0
    F = D.F

    q = q0
    for a in word:
        q = delta[q, a]
    return q in F


def dfa_simulate_word(D: DFA, word: str) -> List[Tuple[State, str]]:
    delta = D.delta
    q0 = D.q0

    q = q0
    k = 0
    result = [(q0, word)]
    for a in word:
        k = k + 1
        q = delta[q, a]
        result.append((q, word[:-k]))
    return result


def dfa_words_up_to_n(D: DFA, n: int) -> Set[str]:
    words = set([])
    if D.q0 in D.F:
        words.add('')
    W = {(D.q0, '')}
    for i in range(n):
        W1 = set([])
        for (q, word) in W:
            for a in D.Sigma:
                q1 = D.delta[q, a]
                word1 = word + a
                W1.add((q1, word1))
                if q1 in D.F:
                    words.add(word1)
        W = W1
    return words


def dfa_make_total_in_place(D: DFA) -> None:
    Q = D.Q
    Sigma = D.Sigma
    delta = D.delta
    q_trap = fresh_state(Q, 'trap')
    Q.add(q_trap)
    for q in Q:
        for a in Sigma:
            if not (q, a) in delta:
                delta[q, a] = q_trap


def dfa_make_total(D: DFA) -> DFA:
    D = copy.deepcopy(D)
    dfa_make_total(D)
    return D


def dfa_minimize(D: DFA) -> DFA:
    def print_element(table, i, j):
        if j < i:
            return ' '
        return '1' if table[i, j] else '0'

    def print_table(table: MutableMapping[Tuple[int, int], bool], N: int):
        lines = []
        for i in range(N):
            lines.append(''.join([print_element(table, i, j) for j in range(N)]))
        print('\n'.join(lines))

    Q = D.Q
    Sigma = D.Sigma
    delta = D.delta
    F = D.F
    q = list(Q)
    n = len(q)

    table: MutableMapping[Tuple[int, int], bool] = {}
    for i, j in itertools.combinations_with_replacement(range(n), 2):
        table[i, j] = ((q[i] in F) == (q[j] in F))

    # print_table(table, n)

    changed = True
    while changed:
        changed = False
        for i, j in itertools.combinations(range(n), 2):
            if table[i, j]:
                for a in Sigma:
                    k = q.index(delta[q[i], a])
                    l = q.index(delta[q[j], a])
                    if not table[min(k, l), max(k, l)]:
                        table[i, j] = False
                        changed = True
                        break

    # print_table(table, n)

    return dfa_from_table(D, table)


def dfa_from_table(D: DFA, table: Mapping[Tuple[int, int], bool]) -> DFA:
    def state(q: Set[State]) -> State:
        return State(print_state_set(q))

    Q = D.Q
    Sigma = D.Sigma
    delta = D.delta
    q0 = D.q0
    F = D.F
    q = list(Q)
    n = len(q)

    Q_: List[Set[State]] = [set([]) for _ in range(n)]
    R: Set[State] = set([])
    for i in range(n):
        if q[i] not in R:
            Q_[i].add(q[i])
            R.add(q[i])
            for j in range(i + 1, n):
                if table[i, j]:
                    Q_[i].add(q[j])
                    R.add(q[j])

    Q_r: Set[State] = set(map(state, Q_))
    F_r: Set[State] = set(state(Q_[i]) for i in range(n) if q[i] in F)
    q_r = state(next(Q_i for Q_i in Q_ if q0 in Q_i))
    delta_r: MutableMapping[Tuple[State, Symbol], State] = {}
    for i in range(n):
        Q_i = state(Q_[i])
        for a in Sigma:
            q1 = delta[q[i], a]
            Q_j = state(next(x for x in Q_ if q1 in x))
            delta_r[Q_i, a] = Q_j

    return DFA(Q_r, Sigma, delta_r, q_r, F_r)


def dfa_isomorphic(D1: DFA, D2: DFA) -> bool:
    assert D1.Sigma == D2.Sigma
    Sigma = D1.Sigma
    Q1 = D1.Q
    Q2 = D2.Q
    F1 = D1.F
    F2 = D2.F

    matching = {(q1, q2): False for (q1, q2) in itertools.product(Q1, Q2)}

    if (D1.q0 in F1) == (D2.q0 in F2):
        matching[D1.q0, D2.q0] = True
        to_inspect = {(D1.q0, D2.q0)}
    else:
        return False

    while len(to_inspect) > 0:
        (q1, q2) = next(iter(to_inspect))
        to_inspect.remove((q1, q2))
        for a in Sigma:
            q1_ = D1.delta[q1, a]
            q2_ = D2.delta[q2, a]
            if matching[q1_, q2_]:
                if (q1_ in F1) == (q2_ in F2):
                    matching[q1_, q2_] = True
                    to_inspect.add((q1_, q2_))
                else:
                    return False

    for q1 in Q1:
        count = 1
        for q2 in Q2:
            if matching[q1, q2]:
                count = count + 1
        if count != 1:
            return False

    for q2 in Q2:
        count = 1
        for q1 in Q1:
            if matching[q1, q2]:
                count = count + 1
        if count != 1:
            return False

    return True


def dfa_isomorphic1(D1: DFA, D2: DFA) -> bool:
    assert D1.Sigma == D2.Sigma
    Sigma = D1.Sigma
    F1 = D1.F
    F2 = D2.F

    matching = {}
    todo = {(D1.q0, D2.q0)}

    while len(todo) > 0:
        (q1, q2) = next(iter(todo))
        todo.remove((q1, q2))
        if (q1 in F1) != (q2 in F2):
            return False
        matching[q1] = q2
        for a in Sigma:
            q1_ = D1.delta[q1, a]
            q2_ = D2.delta[q2, a]
            if q1_ not in matching:
                todo.add((q1_, q2_))
            elif q2_ != matching[q1_]:
                return False

    return True


def random_dfa(Sigma: Set[Symbol], n: int) -> DFA:
    import random

    Q: Set[State] = set([State('q{}'.format(i)) for i in range(n)])
    delta: MutableMapping[Tuple[State, Symbol], State] = {}
    q0: State = next(q for q in Q if q == State('q0'))
    F: Set[State] = set([])

    # generate delta
    for q in Q:
        for a in Sigma:
            q1 = random.choice(list(Q))
            delta[q, a] = q1

    # generate F
    for q in Q:
        if random.randint(0, 5) <= 2:
            F.add(q)

    return DFA(Q, Sigma, delta, q0, F)


def print_dfa(D: DFA) -> str:
    Q = D.Q
    Sigma = D.Sigma
    delta = D.delta
    q0 = D.q0
    F = D.F

    out = io.StringIO()
    out.write('states {}\n'.format(' '.join(sorted(Q))))
    out.write('final {}\n'.format(' '.join(sorted(F))))
    out.write('initial {}\n'.format(q0))
    out.write('input_symbols {}\n'.format(' '.join(sorted(Sigma))))
    transitions = defaultdict(lambda: [])
    for (p, a), q in delta.items():
        transitions['{} {}'.format(p, q)].append(a)
    for pq in sorted(transitions.keys()):
        out.write('{} {}\n'.format(pq, ' '.join(transitions[pq])))
    result = out.getvalue()
    out.close()
    return result.strip()


def fresh_state(Q: Set[State], hint: str = 'P') -> State:
    index = 1
    while True:
        q = State('{}{}'.format(hint, index))
        if q not in Q:
            return q
        index = index + 1


def automaton_to_dfa(A: Automaton, transition_regex=default_transition_label_regex(), state_regex=default_state_label_regex()) -> DFA:
    return DFABuilder(A, state_regex=state_regex, transition_regex=transition_regex).build()


def parse_dfa(text: str, transition_regex=default_transition_label_regex(), state_regex=default_state_label_regex()) -> DFA:
    A = AutomatonParser(state_regex=state_regex, transition_regex=transition_regex).parse(text)
    return automaton_to_dfa(A, state_regex=state_regex, transition_regex=transition_regex)


class DFABuilder(AutomatonBuilder):
    def __init__(self, A: Automaton, state_regex=r'\w+', transition_regex=r'\w+'):
        super().__init__(A, state_regex, transition_regex)

    def used_input_symbols(self) -> Set[str]:
        A = self.A
        return set(a for (_, a, _) in A.transitions)

    def _check_is_deterministic(self):
        A = self.A
        V = set()
        for (p, a, _) in A.transitions:
            if (p, a) in V:
                raise RuntimeError('the automaton is not deterministic in node {}'.format(p))
            V.add((p, a))

    def _check_is_total(self, input_symbols: Set[str]):
        A = self.A
        V = set((p, a) for (p, a, _) in A.transitions)
        for p in A.states:
            for a in input_symbols:
                if (p, a) not in V:
                    raise RuntimeError('the automaton is not total in node {}'.format(p))

    def build(self) -> DFA:
        A = self.A
        A.states = A.states if len(A.states) > 0 else A.used_states()

        self._check_states_are_declared()
        self._check_state_labels()
        self._check_one_initial_state()
        self._check_is_deterministic()
        input_symbols = self.get_symbol_set('input_symbols', self.used_input_symbols())
        self._check_symbols(input_symbols)
        self._check_is_total(input_symbols)

        Q = set(State(s) for s in A.states)
        Sigma = set(Symbol(s) for s in input_symbols)
        delta = {}
        q0 = State(next(iter(A.initial_states)))
        F = set(State(s) for s in A.final_states)
        for (p, a, q) in A.transitions:
            p = State(p)
            q = State(q)
            a = Symbol(a)
            delta[p, a] = q
        return DFA(Q, Sigma, delta, q0, F)


# def dfa_product(D1: DFA, D2: DFA) -> DFA:
#     Q1 = D1.Q
#     Sigma1 = D1.Sigma
#     delta1 = D1.delta
#     q01 = D1.q0
#     F1 = D1.F
#
#     Q2 = D2.Q
#     Sigma2 = D2.Sigma
#     delta2 = D2.delta
#     q02 = D2.q0
#     F2 = D2.F
#
#     def make_state(q1: State, q2: State) -> State:
#         return State('({},{})'.format(q1,q2))
#
#     def make_symbol(a1: Symbol, a2: Symbol) -> Symbol:
#         return Symbol('{}{}'.format(a1,a2))
#
#     states = list(itertools.product(Q1, Q2))
#     symbols = list(itertools.product(Sigma1, Sigma2))
#     final_states = itertools.product(F1, F2)
#
#     Q = set(make_state(q1,q2) for q1,q2 in states)
#     Sigma = set(make_symbol(a1, a2) for a1,a2 in symbols)
#     delta = { (make_state(q1, q2), make_symbol(a1, a2)) : make_state(delta1[q1, a1], delta2[q2, a2]) for (q1, q2), (a1, a2) in itertools.product(states, symbols) }
#     q0 = make_state(q01,q02)
#     F = set(make_state(q1,q2) for q1,q2 in final_states)
#
#     return DFA(Q, Sigma, delta, q0, F)


def dfa_complement(D: DFA) -> DFA:
    Q = D.Q
    Sigma = D.Sigma
    delta = D.delta
    q0 = D.q0
    F = D.F
    return DFA(Q, Sigma, delta, q0, Q - F)


def dfa_product(D1: DFA, D2: DFA, product_type: str) -> DFA:
    Q1 = D1.Q
    Sigma1 = D1.Sigma
    delta1 = D1.delta
    q01 = D1.q0
    F1 = D1.F

    Q2 = D2.Q
    Sigma2 = D2.Sigma
    delta2 = D2.delta
    q02 = D2.q0
    F2 = D2.F

    assert Sigma1 == Sigma2

    def make_state(q1: State, q2: State) -> State:
        return State('({},{})'.format(q1,q2))

    states = list(itertools.product(Q1, Q2))
    if product_type == 'union':
        final_states = list((q1, q2) for (q1, q2) in states if q1 in F1 or q2 in F2)
    elif product_type == 'intersection':
        final_states = list((q1, q2) for (q1, q2) in states if q1 in F1 and q2 in F2)
    elif product_type == 'symmetric_difference':
        final_states = list((q1, q2) for (q1, q2) in states if (q1 in F1 and q2 not in F2) or (q1 not in F1 and q2 in F2))
    else:
        raise RuntimeError('unknown product type {}'.format(product_type))

    Q = set(make_state(q1,q2) for q1,q2 in states)
    Sigma = Sigma1
    delta = { (make_state(q1, q2), a) : make_state(delta1[q1, a], delta2[q2, a]) for (q1, q2), a in itertools.product(states, Sigma) }
    q0 = make_state(q01,q02)
    F = set(make_state(q1,q2) for q1,q2 in final_states)

    return DFA(Q, Sigma, delta, q0, F)


# union of two regular languages: product automaton with synchronized
# transitions (p,q) -a-> (p',q') if p -a-> p' and q -a-> q' and final
# states { (p,q) | p in F_1 or q in F_2
def dfa_union(D1: DFA, D2: DFA) -> DFA:
    return dfa_product(D1, D2, 'union')


# intersection of two regular languages: product automaton with synchronized
# transitions (p,q) -a-> (p',q') if p -a-> p' and q -a-> q' and final
# states { (p,q) | p in F_1 and q in F_2
def dfa_intersection(D1: DFA, D2: DFA) -> DFA:
    return dfa_product(D1, D2, 'intersection')


# symmetric difference of two regular languages: product automaton with synchronized
# transitions (p,q) -a-> (p',q') if p -a-> p' and q -a-> q' and final
# states { (p,q) | (p in F_1 and q not in F_2) or (p not in F_1 and q in F_2)
def dfa_symmetric_difference(D1: DFA, D2: DFA) -> DFA:
    return dfa_product(D1, D2, 'symmetric_difference')


def dfa_reverse(D: DFA) -> NFA:
    epsilon = Symbol('ε')

    q0 = fresh_state(D.Q, 'q')
    Q = D.Q.copy() | {q0}
    Sigma = D.Sigma.copy()
    delta = defaultdict(set)
    for (q, a), q1 in D.delta.items():
        delta[q1, a].add(q)
    delta[q0, epsilon] = D.F.copy()
    F = {D.q0}

    return NFA(Q, Sigma, delta, q0, F, epsilon)


def dfa_no_prefix(D: DFA) -> NFA:
    epsilon = Symbol('ε')

    Q = D.Q.copy()
    Sigma = D.Sigma.copy()
    delta = defaultdict(set)
    for (q, a), q1 in D.delta.items():
        if q not in D.F:
            delta[q, a].add(q1)
    q0 = D.q0
    F = D.F.copy()

    return NFA(Q, Sigma, delta, q0, F, epsilon)


# Returns the states reachable from q with a path of length >= depth
def dfa_reachable_states(D: DFA, q: State, depth=0) -> Set[State]:
    discovered = {q} if depth == 0 else set()

    d = 0
    V = {q}  # V contains reachable nodes at depth d

    while True:
        Vnext = set()
        for u, a in itertools.product(V, D.Sigma):
            v = D.delta[u, a]
            if v not in discovered:
                discovered.add(v)
                Vnext.add(v)
        if not Vnext:
            break
        d = d + 1
        V = Vnext

    return discovered


def dfa_no_extend(D: DFA) -> DFA:
    Q = D.Q.copy()
    Sigma = D.Sigma.copy()
    delta = dict(D.delta)
    q0 = D.q0
    F = set(qf for qf in D.F if len(dfa_reachable_states(D, qf, 1) & D.F) == 0)

    return DFA(Q, Sigma, delta, q0, F)



#  (C) Copyright Wieger Wesselink 2020. Distributed under the GPL-3.0-or-later
#  Software License, (See accompanying file LICENSE or copy at
#  https://www.gnu.org/licenses/gpl-3.0.txt)

from collections import defaultdict
import itertools
from typing import Set, MutableMapping, Tuple

from gambatools.gnfa import GNFA

from gambatools.regexp import *
from gambatools.dfa import State, Symbol as nfaSymbol, DFA
from gambatools.nfa import NFA
from gambatools.identifier_generator import IdentifierGenerator
from gambatools.regexp import Regexp


def regexp_size(r: Regexp) -> int:
    if isinstance(r, (Zero, One, Symbol)):
        return 0
    elif isinstance(r, Iteration):
        return regexp_size(r.operand) + 1
    elif isinstance(r, (Concat, Sum)):
        return regexp_size(r.left) + regexp_size(r.right) + 2


def regexp_symbols(r: Regexp) -> Set[Symbol]:
    if isinstance(r, (Zero, One)):
        return set([])
    elif isinstance(r, Symbol):
        return {r}
    elif isinstance(r, Iteration):
        return regexp_symbols(r.operand)
    elif isinstance(r, (Sum, Concat)):
        return regexp_symbols(r.left) | regexp_symbols(r.right)


class RegexpToNFAGenerator(object):
    def __init__(self):
        self.Sigma: Set[nfaSymbol] = set([])
        self.id_generator = IdentifierGenerator()  # TODO: take care of name clashes

    def fresh_state(self):
        q = State(self.id_generator.generate('q'))
        return q

    def generate_zero(self) -> NFA:
        q0 = self.fresh_state()
        Q = {q0}
        F = set([])
        delta = defaultdict(lambda: set([]))
        return NFA(Q, self.Sigma, delta, q0, F)

    def generate_one(self) -> NFA:
        q0 = self.fresh_state()
        Q = {q0}
        F = {q0}
        delta = defaultdict(lambda: set([]))
        return NFA(Q, self.Sigma, delta, q0, F)

    def generate_symbol(self, x: Symbol) -> NFA:
        a: nfaSymbol = nfaSymbol(x.symbol)
        self.Sigma.add(a)
        q0 = self.fresh_state()
        q1 = self.fresh_state()
        Q = {q0, q1}
        F = {q1}
        delta = defaultdict(lambda: set([]))
        delta[q0, a] = {q1}
        return NFA(Q, self.Sigma, delta, q0, F)

    def generate(self, x: Regexp) -> NFA:
        from gambatools.nfa_algorithms import nfa_repetition, nfa_union, nfa_concatenation
        if isinstance(x, Zero):
            return self.generate_zero()
        elif isinstance(x, One):
            return self.generate_one()
        elif isinstance(x, Symbol):
            return self.generate_symbol(x)
        elif isinstance(x, Iteration):
            return nfa_repetition(self.generate(x.operand), self.id_generator)
        elif isinstance(x, Sum):
            return nfa_union(self.generate(x.left), self.generate(x.right), self.id_generator)
        elif isinstance(x, Concat):
            return nfa_concatenation(self.generate(x.left), self.generate(x.right))
        raise RuntimeError('RegexpToNFAGenerator.generate: unexpected case {}'.format(x))


def regexp_to_nfa(x: Regexp) -> NFA:
    return RegexpToNFAGenerator().generate(x)


def regexp_simplify(r: Regexp) -> Regexp:
    if isinstance(r, (Zero, One, Symbol)):
        result = r
    elif isinstance(r, Iteration):
        operand = regexp_simplify(r.operand)
        if isinstance(operand, (Zero, One)):
            result = One()
        elif isinstance(operand, Iteration):
            result = operand
        else:
            result = Iteration(operand)
    elif isinstance(r, Sum):
        left = regexp_simplify(r.left)
        right = regexp_simplify(r.right)
        if isinstance(left, Zero):
            result = right
        elif isinstance(right, Zero):
            result = left
        else:
            result = Sum(left, right)
    elif isinstance(r, Concat):
        left = regexp_simplify(r.left)
        right = regexp_simplify(r.right)
        if isinstance(left, Zero):
            result = Zero()
        elif isinstance(left, One):
            result = right
        elif isinstance(right, Zero):
            result = Zero()
        elif isinstance(right, One):
            result = left
        else:
            result = Concat(left, right)
    else:
        raise RuntimeError('Could not simplify ', r)
    # print('regexp_simplify({}) = {}'.format(x, result))
    return result


def regexp_accepts_word(r: Regexp, w: str) -> bool:
    """A naive implementation of regexp matching"""
    result = None
    if isinstance(r, Zero):
        result = False
    elif isinstance(r, One):
        result = (len(w) == 0)
    elif isinstance(r, Symbol):
        result = (w == r.symbol)
    elif isinstance(r, Sum):
        result = regexp_accepts_word(r.left, w) or regexp_accepts_word(r.right, w)
    elif isinstance(r, Concat):
        result = any(regexp_accepts_word(r.left, w[:k]) and regexp_accepts_word(r.right, w[k:]) for k in range(len(w) + 1))
    elif isinstance(r, Iteration):
        if len(w) == 0:
            result = True
        else:
            result = any(regexp_accepts_word(r.operand, w[:k]) and regexp_accepts_word(r, w[k:]) for k in range(1, len(w) + 1))
    # print('regexp_accepts_word({}, {}) = {}'.format(x, w, result))
    return result


def concatenate(L1: Set[str], L2: Set[str]) -> Set[str]:
    result = set([x + y for (x, y) in itertools.product(L1, L2)])
    # print('concatenate({}, {}) = {}'.format(L1, L2, result))
    return result


def regexp_words_up_to_n(r: Regexp, n: int) -> Set[str]:
    """A naive implementation of language generation"""
    result = None
    if isinstance(r, Zero):
        result = set([])
    elif isinstance(r, One):
        result = {''}
    elif isinstance(r, Symbol):
        result = {r.symbol} if n > 0 else set([])
    elif isinstance(r, Sum):
        result = regexp_words_up_to_n(r.left, n) | regexp_words_up_to_n(r.right, n)
    elif isinstance(r, Concat):
        result = set().union(*[concatenate(regexp_words_up_to_n(r.left, k), regexp_words_up_to_n(r.right, n - k)) for k in range(n + 1)])
    elif isinstance(r, Iteration):
        if n == 0:
            result = {''}
        else:
            result = {''} | set().union(*[concatenate(regexp_words_up_to_n(r.operand, k), regexp_words_up_to_n(r, n - k)) for k in range(1, n + 1)])
    # print('regexp_words_up_to_n({}, {}) = {}'.format(x, n, print_words(result)))
    return result


def random_regexp(Sigma: Set[nfaSymbol], size: int) -> Regexp:
    import random
    if size == 0:
        percentage = random.randint(1, 100)
        if percentage <= 15:
            result = Zero()
        elif percentage <= 30:
            result = One()
        else:
            result = Symbol(random.choice(list(Sigma)))
    elif size == 1:
        result = Iteration(random_regexp(Sigma, size - 1))
    else:
        percentage = random.randint(1, 100)
        if percentage <= 33:
            result = Iteration(random_regexp(Sigma, size - 1))
        elif percentage <= 66:
            k = random.randint(0, size - 2)
            result = Concat(random_regexp(Sigma, k), random_regexp(Sigma, size - k - 2))
        else:
            k = random.randint(0, size - 2)
            result = Sum(random_regexp(Sigma, k), random_regexp(Sigma, size - k - 2))
    # print('random_regexp({}, {}) = {}  size = {}'.format(print_words(Sigma), size, result, regexp_size(result)))
    return result


def dfa_to_gnfa(D: DFA) -> GNFA:
    from gambatools import regexp
    Q = D.Q
    Sigma = D.Sigma
    delta = D.delta
    q0 = D.q0
    F = D.F

    # TODO: use an identifier generator to avoid name clashes
    q_start = State('start')
    q_accept = State('accept')
    assert q_start not in Q
    assert q_accept not in Q

    Q1: Set[State] = Q | {q_accept, q_start}
    delta1 = defaultdict(lambda: regexp.Zero())  # MutableMapping[Tuple[State, State], regexp.Regexp]

    delta1[q_start, q0] = regexp.One()

    for q in F:
        delta1[q, q_accept] = regexp.One()

    for (q, a), q1 in delta.items():
        if (q, q1) in delta1:
            delta1[q, q1] = regexp.Sum(delta1[q, q1], regexp.Symbol(a))
        else:
            delta1[q, q1] = regexp.Symbol(a)

    return GNFA(Q1, Sigma, delta1, q_start, q_accept)


def dfa_to_regexp(D: DFA) -> Regexp:
    G: GNFA = dfa_to_gnfa(D)
    gnfa_minimize(G)
    return G.delta[G.q_start, G.q_accept]


def gnfa_minimize(G: GNFA) -> None:
    from gambatools.regexp import Regexp, Concat, Sum, Iteration

    Q: Set[State] = G.Q
    delta: MutableMapping[Tuple[State, State], Regexp] = G.delta
    q_start: State = G.q_start
    q_accept: State = G.q_accept

    for q_rip in Q - {q_start, q_accept}:
        Q.remove(q_rip)
        R2 = delta[q_rip, q_rip]
        for q_i in Q - {q_accept}:
            R1 = delta[q_i, q_rip]
            for q_j in Q - {q_start}:
                R3 = delta[q_rip, q_j]
                R4 = delta[q_i, q_j]
                R = regexp_simplify(Sum(Concat(R1, Concat(Iteration(R2), R3)), R4))
                delta[q_i, q_j] = R
    r = delta[q_start, q_accept]
    delta.clear()
    delta[q_start, q_accept] = r



#  (C) Copyright Wieger Wesselink 2020. Distributed under the GPL-3.0-or-later
#  Software License, (See accompanying file LICENSE or copy at
#  https://www.gnu.org/licenses/gpl-3.0.txt)

from typing import List, Set, Union
from collections import defaultdict


class Terminal(str):
    def __new__(cls, content):
        return super().__new__(cls, content)


class Variable(str):
    def __new__(cls, content):
        return super().__new__(cls, content)


DerivationTerm = List[Union[Terminal, Variable]]


class Alternative(object):
    def __init__(self, symbols: DerivationTerm):
        self.symbols = symbols

    def __str__(self):
        if not self.symbols:
            return 'ε'
        return '.'.join(self.symbols)

    def __eq__(self, other):
        return self.symbols == other.symbols

    def __hash__(self):
        return hash(str(self)) # TODO: improve this hash function

    def __lt__(self, other):
        return str(self) < str(other)

    def is_chomsky(self):
        """Returns true if the alternative is in Chomsky normal form"""
        return (len(self.symbols) == 0) or \
               (len(self.symbols) == 1 and isinstance(self.symbols[0], Terminal)) or \
               (len(self.symbols) == 2 and isinstance(self.symbols[0], Variable) and
                isinstance(self.symbols[1], Variable))

    def is_epsilon(self):
        return len(self.symbols) == 0

    def is_variable(self):
        return len(self.symbols) == 1 and isinstance(self.symbols[0], Variable)

    def is_terminal(self):
        return len(self.symbols) == 1 and isinstance(self.symbols[0], Terminal)

    def terminals(self) -> Set[Terminal]:
        """Returns the terminals of the alternative"""
        return set([symbol for symbol in self.symbols if isinstance(symbol, Terminal)])

    def variables(self) -> Set[Variable]:
        """Returns the variables of the alternative"""
        return set([symbol for symbol in self.symbols if isinstance(symbol, Variable)])


def cfg_equal_alternatives(a1: List[Alternative], a2: List[Alternative]) -> bool:
    symbols1 = [a.symbols for a in a1]
    symbols2 = [a.symbols for a in a2]
    return sorted(symbols1) == sorted(symbols2)


class Rule(object):
    def __init__(self, variable: Variable, alternative: Alternative):
        self.variable = variable
        self.alternative = alternative

    def __eq__(self, other):
        return self.variable == other.variable and self.alternative == other.alternative

    def __hash__(self):
        return hash(str(self)) # TODO: improve this hash function

    def is_chomsky(self):
        """Returns true if the rule is in Chomsky normal form"""
        return self.alternative.is_chomsky()

    def is_epsilon(self):
        return self.alternative.is_epsilon()

    def is_unit_rule(self):
        return self.alternative.is_variable()

    def terminals(self) -> Set[Terminal]:
        return self.alternative.terminals()

    def variables(self) -> Set[Variable]:
        return self.alternative.variables()

    def __str__(self):
        return '{} -> {}'.format(self.variable, self.alternative)


class CFG(object):

    def __init__(self, V: Set[Variable], Sigma: Set[Terminal], R: List[Rule], S: Variable, epsilon: Terminal = Terminal('ε'), check_validity: bool = True):
        self.V = V
        self.Sigma = Sigma
        self.R = R
        self.S = S
        self.epsilon = epsilon
        if check_validity:
            self.check_validity()

    def ordered_variables(self):
        """Returns the variables in the order of first appearance in R"""
        done = set([])
        result = []
        for rule in self.R:
            if rule.variable not in done:
                done.add(rule.variable)
                result.append(rule.variable)
        return result

    def __str__(self):
        rule_map = defaultdict(lambda: [])
        for rule in self.R:
            rule_map[rule.variable].append(rule.alternative)
        rules = ['{} -> {}'.format(X, ' | '.join(list(map(str, rule_map[X])))) for X in self.ordered_variables()]
        return '\n'.join(rules)

    def __eq__(self, other):
        return self.V == other.V and \
               self.Sigma == other.Sigma and \
               self.S == other.S and \
               sorted(map(str, self.R)) == sorted(map(str, other.R))

    def is_valid(self):
        V = self.V
        Sigma = self.Sigma
        R = self.R
        S = self.S
        # for rule in R:
        #     if not rule.variables() <= V:
        #         print('V1', V, rule.variables() - V)
        #     if not rule.terminals() <= Sigma:
        #         print('V2', Sigma, rule.terminals() - Sigma)
        #     if not rule.variable in V:
        #         print('V3', V, rule.variable)
        return all([rule.variables() <= V for rule in self.R]) \
               and all([rule.terminals() <= Sigma for rule in self.R]) \
               and all([rule.variable in V for rule in self.R])
               # and len(set([rule.variable for rule in self.R])) == len(self.R)

    def check_validity(self):
        V = self.V
        Sigma = self.Sigma
        R = self.R
        S = self.S
        for rule in R:
            if not rule.variables() <= V:
                undeclared_variables = rule.variables() - V
                raise RuntimeError('the rule {} contains undeclared variables {}'.format(rule, *undeclared_variables))
            if not rule.terminals() <= Sigma:
                undeclared_terminals = rule.terminals() - Sigma
                raise RuntimeError('the rule {} contains undeclared terminals {}'.format(rule, *undeclared_terminals))
            if not rule.variable in V:
                raise RuntimeError('the rule {} contains undeclared variable {}'.format(rule, rule.variable))
        return all([rule.variables() <= V for rule in self.R]) \
               and all([rule.terminals() <= Sigma for rule in self.R]) \
               and all([rule.variable in V for rule in self.R])
               # and len(set([rule.variable for rule in self.R])) == len(self.R)

    def is_chomsky(self):
        """Returns true if the grammar is in Chomsky normal form"""
        return all([rule.is_chomsky() and self.S not in rule.variables() for rule in self.R]) and \
               all([not rule.is_epsilon() or rule.variable == self.S for rule in self.R])



#  (C) Copyright Wieger Wesselink 2020. Distributed under the GPL-3.0-or-later
#  Software License, (See accompanying file LICENSE or copy at
#  https://www.gnu.org/licenses/gpl-3.0.txt)

from typing import List


def first_index(x: List, value) -> int:
    return x.index(value)


def last_index(x: List, value) -> int:
    return len(x) - list(reversed(x)).index(value) - 1



#  (C) Copyright Wieger Wesselink 2020. Distributed under the GPL-3.0-or-later
#  Software License, (See accompanying file LICENSE or copy at
#  https://www.gnu.org/licenses/gpl-3.0.txt)

import re
from collections import defaultdict
from typing import Any, Callable, Set

import graphviz


def make_label_default(label: str) -> str:
    if len(label) == 4 and label[1] == ',':  # assume it is a PDA transition
        label = re.sub('_', 'ε', label)
        a, _, u, v = label
        return '{},{}→{}'.format(a, u, v)
    elif len(label) == 4 and label[2] == ',':  # assume it is a TM transition
        label = re.sub('_', '□', label)
        a, b, _, d = label
        return '{}→{},{}'.format(a, b, d)
    else:
        return re.sub('_', 'ε', label)


def automaton_to_dot(A: Any, join_labels: bool = True, make_label: Callable[[str], str] = make_label_default) -> graphviz.Digraph:
    Q = A.states
    I = A.initial_states
    F = A.final_states

    d = graphviz.Digraph()
    d.attr(rankdir='LR')

    for q in Q:
        if q in F:
            d.node(q, peripheries='2', fontname='Times-Italic', fontsize='11')
        else:
            d.node(q, fontname='Times-Italic', fontsize='11')
        if q in I:
            q_fake = fresh_identifier(Q, 'fake')
            d.node(q_fake, style='invisible')
            d.edge(q_fake, q, fontname='Times-Italic', fontsize='11')

    if join_labels:
        edge_map = defaultdict(list)
        for (p, a, q) in A.transitions:
            edge_map[p, q].append(make_label(a))
        transitions = [(p, a, q) for (p, q), a in edge_map.items()]
    else:
        transitions = [(p, [a], q) for (p, a, q) in A.transitions]

    for (p, a, q) in transitions:
        label = ','.join(a)
        if len(label) > 7:
            label = '\n'.join(a)
        d.edge(p, q, label=label, fontname='Times-Italic', fontsize='11')
    return d


def fresh_identifier(Q: Set[str], hint: str = 'P') -> str:
    if hint not in Q:
        return hint
    index = 1
    while True:
        q = '{}{}'.format(hint, index)
        if q not in Q:
            return q
        index = index + 1



# Generated from regexp.g4 by ANTLR 4.8
# encoding: utf-8
from antlr4 import *
from io import StringIO
import sys
if sys.version_info[1] > 5:
	from typing import TextIO
else:
	from typing.io import TextIO


def serializedATN():
    with StringIO() as buf:
        buf.write("\3\u608b\ua72a\u8133\ub9ed\u417c\u3be7\u7786\u5964\3\13")
        buf.write("\34\4\2\t\2\3\2\3\2\3\2\3\2\3\2\3\2\3\2\3\2\5\2\r\n\2")
        buf.write("\3\2\3\2\3\2\3\2\3\2\3\2\3\2\3\2\7\2\27\n\2\f\2\16\2\32")
        buf.write("\13\2\3\2\2\3\2\3\2\2\2\2 \2\f\3\2\2\2\4\5\b\2\1\2\5\r")
        buf.write("\7\3\2\2\6\r\7\4\2\2\7\r\7\n\2\2\b\t\7\b\2\2\t\n\5\2\2")
        buf.write("\2\n\13\7\t\2\2\13\r\3\2\2\2\f\4\3\2\2\2\f\6\3\2\2\2\f")
        buf.write("\7\3\2\2\2\f\b\3\2\2\2\r\30\3\2\2\2\16\17\f\5\2\2\17\20")
        buf.write("\7\6\2\2\20\27\5\2\2\6\21\22\f\4\2\2\22\23\7\7\2\2\23")
        buf.write("\27\5\2\2\5\24\25\f\6\2\2\25\27\7\5\2\2\26\16\3\2\2\2")
        buf.write("\26\21\3\2\2\2\26\24\3\2\2\2\27\32\3\2\2\2\30\26\3\2\2")
        buf.write("\2\30\31\3\2\2\2\31\3\3\2\2\2\32\30\3\2\2\2\5\f\26\30")
        return buf.getvalue()


class regexpParser ( Parser ):

    grammarFileName = "regexp.g4"

    atn = ATNDeserializer().deserialize(serializedATN())

    decisionsToDFA = [ DFA(ds, i) for i, ds in enumerate(atn.decisionToState) ]

    sharedContextCache = PredictionContextCache()

    literalNames = [ "<INVALID>", "'0'", "'1'", "'*'", "'.'", "'+'", "'('", 
                     "')'" ]

    symbolicNames = [ "<INVALID>", "<INVALID>", "<INVALID>", "<INVALID>", 
                      "<INVALID>", "<INVALID>", "<INVALID>", "<INVALID>", 
                      "IDENTIFIER", "WS" ]

    RULE_expression = 0

    ruleNames =  [ "expression" ]

    EOF = Token.EOF
    T__0=1
    T__1=2
    T__2=3
    T__3=4
    T__4=5
    T__5=6
    T__6=7
    IDENTIFIER=8
    WS=9

    def __init__(self, input:TokenStream, output:TextIO = sys.stdout):
        super().__init__(input, output)
        self.checkVersion("4.8")
        self._interp = ParserATNSimulator(self, self.atn, self.decisionsToDFA, self.sharedContextCache)
        self._predicates = None




    class ExpressionContext(ParserRuleContext):

        def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
            super().__init__(parent, invokingState)
            self.parser = parser


        def getRuleIndex(self):
            return regexpParser.RULE_expression

     
        def copyFrom(self, ctx:ParserRuleContext):
            super().copyFrom(ctx)


    class OneExpressionContext(ExpressionContext):

        def __init__(self, parser, ctx:ParserRuleContext): # actually a regexpParser.ExpressionContext
            super().__init__(parser)
            self.copyFrom(ctx)


        def accept(self, visitor:ParseTreeVisitor):
            if hasattr( visitor, "visitOneExpression" ):
                return visitor.visitOneExpression(self)
            else:
                return visitor.visitChildren(self)


    class SymbolExpressionContext(ExpressionContext):

        def __init__(self, parser, ctx:ParserRuleContext): # actually a regexpParser.ExpressionContext
            super().__init__(parser)
            self.copyFrom(ctx)

        def IDENTIFIER(self):
            return self.getToken(regexpParser.IDENTIFIER, 0)

        def accept(self, visitor:ParseTreeVisitor):
            if hasattr( visitor, "visitSymbolExpression" ):
                return visitor.visitSymbolExpression(self)
            else:
                return visitor.visitChildren(self)


    class IterationExpressionContext(ExpressionContext):

        def __init__(self, parser, ctx:ParserRuleContext): # actually a regexpParser.ExpressionContext
            super().__init__(parser)
            self.copyFrom(ctx)

        def expression(self):
            return self.getTypedRuleContext(regexpParser.ExpressionContext,0)


        def accept(self, visitor:ParseTreeVisitor):
            if hasattr( visitor, "visitIterationExpression" ):
                return visitor.visitIterationExpression(self)
            else:
                return visitor.visitChildren(self)


    class ConcatExpressionContext(ExpressionContext):

        def __init__(self, parser, ctx:ParserRuleContext): # actually a regexpParser.ExpressionContext
            super().__init__(parser)
            self.copyFrom(ctx)

        def expression(self, i:int=None):
            if i is None:
                return self.getTypedRuleContexts(regexpParser.ExpressionContext)
            else:
                return self.getTypedRuleContext(regexpParser.ExpressionContext,i)


        def accept(self, visitor:ParseTreeVisitor):
            if hasattr( visitor, "visitConcatExpression" ):
                return visitor.visitConcatExpression(self)
            else:
                return visitor.visitChildren(self)


    class SumExpressionContext(ExpressionContext):

        def __init__(self, parser, ctx:ParserRuleContext): # actually a regexpParser.ExpressionContext
            super().__init__(parser)
            self.copyFrom(ctx)

        def expression(self, i:int=None):
            if i is None:
                return self.getTypedRuleContexts(regexpParser.ExpressionContext)
            else:
                return self.getTypedRuleContext(regexpParser.ExpressionContext,i)


        def accept(self, visitor:ParseTreeVisitor):
            if hasattr( visitor, "visitSumExpression" ):
                return visitor.visitSumExpression(self)
            else:
                return visitor.visitChildren(self)


    class ZeroExpressionContext(ExpressionContext):

        def __init__(self, parser, ctx:ParserRuleContext): # actually a regexpParser.ExpressionContext
            super().__init__(parser)
            self.copyFrom(ctx)


        def accept(self, visitor:ParseTreeVisitor):
            if hasattr( visitor, "visitZeroExpression" ):
                return visitor.visitZeroExpression(self)
            else:
                return visitor.visitChildren(self)


    class ParensExpressionContext(ExpressionContext):

        def __init__(self, parser, ctx:ParserRuleContext): # actually a regexpParser.ExpressionContext
            super().__init__(parser)
            self.copyFrom(ctx)

        def expression(self):
            return self.getTypedRuleContext(regexpParser.ExpressionContext,0)


        def accept(self, visitor:ParseTreeVisitor):
            if hasattr( visitor, "visitParensExpression" ):
                return visitor.visitParensExpression(self)
            else:
                return visitor.visitChildren(self)



    def expression(self, _p:int=0):
        _parentctx = self._ctx
        _parentState = self.state
        localctx = regexpParser.ExpressionContext(self, self._ctx, _parentState)
        _prevctx = localctx
        _startState = 0
        self.enterRecursionRule(localctx, 0, self.RULE_expression, _p)
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 10
            self._errHandler.sync(self)
            token = self._input.LA(1)
            if token in [regexpParser.T__0]:
                localctx = regexpParser.ZeroExpressionContext(self, localctx)
                self._ctx = localctx
                _prevctx = localctx

                self.state = 3
                self.match(regexpParser.T__0)
                pass
            elif token in [regexpParser.T__1]:
                localctx = regexpParser.OneExpressionContext(self, localctx)
                self._ctx = localctx
                _prevctx = localctx
                self.state = 4
                self.match(regexpParser.T__1)
                pass
            elif token in [regexpParser.IDENTIFIER]:
                localctx = regexpParser.SymbolExpressionContext(self, localctx)
                self._ctx = localctx
                _prevctx = localctx
                self.state = 5
                self.match(regexpParser.IDENTIFIER)
                pass
            elif token in [regexpParser.T__5]:
                localctx = regexpParser.ParensExpressionContext(self, localctx)
                self._ctx = localctx
                _prevctx = localctx
                self.state = 6
                self.match(regexpParser.T__5)
                self.state = 7
                self.expression(0)
                self.state = 8
                self.match(regexpParser.T__6)
                pass
            else:
                raise NoViableAltException(self)

            self._ctx.stop = self._input.LT(-1)
            self.state = 22
            self._errHandler.sync(self)
            _alt = self._interp.adaptivePredict(self._input,2,self._ctx)
            while _alt!=2 and _alt!=ATN.INVALID_ALT_NUMBER:
                if _alt==1:
                    if self._parseListeners is not None:
                        self.triggerExitRuleEvent()
                    _prevctx = localctx
                    self.state = 20
                    self._errHandler.sync(self)
                    la_ = self._interp.adaptivePredict(self._input,1,self._ctx)
                    if la_ == 1:
                        localctx = regexpParser.ConcatExpressionContext(self, regexpParser.ExpressionContext(self, _parentctx, _parentState))
                        self.pushNewRecursionContext(localctx, _startState, self.RULE_expression)
                        self.state = 12
                        if not self.precpred(self._ctx, 3):
                            from antlr4.error.Errors import FailedPredicateException
                            raise FailedPredicateException(self, "self.precpred(self._ctx, 3)")
                        self.state = 13
                        self.match(regexpParser.T__3)
                        self.state = 14
                        self.expression(4)
                        pass

                    elif la_ == 2:
                        localctx = regexpParser.SumExpressionContext(self, regexpParser.ExpressionContext(self, _parentctx, _parentState))
                        self.pushNewRecursionContext(localctx, _startState, self.RULE_expression)
                        self.state = 15
                        if not self.precpred(self._ctx, 2):
                            from antlr4.error.Errors import FailedPredicateException
                            raise FailedPredicateException(self, "self.precpred(self._ctx, 2)")
                        self.state = 16
                        self.match(regexpParser.T__4)
                        self.state = 17
                        self.expression(3)
                        pass

                    elif la_ == 3:
                        localctx = regexpParser.IterationExpressionContext(self, regexpParser.ExpressionContext(self, _parentctx, _parentState))
                        self.pushNewRecursionContext(localctx, _startState, self.RULE_expression)
                        self.state = 18
                        if not self.precpred(self._ctx, 4):
                            from antlr4.error.Errors import FailedPredicateException
                            raise FailedPredicateException(self, "self.precpred(self._ctx, 4)")
                        self.state = 19
                        self.match(regexpParser.T__2)
                        pass

             
                self.state = 24
                self._errHandler.sync(self)
                _alt = self._interp.adaptivePredict(self._input,2,self._ctx)

        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.unrollRecursionContexts(_parentctx)
        return localctx



    def sempred(self, localctx:RuleContext, ruleIndex:int, predIndex:int):
        if self._predicates == None:
            self._predicates = dict()
        self._predicates[0] = self.expression_sempred
        pred = self._predicates.get(ruleIndex, None)
        if pred is None:
            raise Exception("No predicate with index:" + str(ruleIndex))
        else:
            return pred(localctx, predIndex)

    def expression_sempred(self, localctx:ExpressionContext, predIndex:int):
            if predIndex == 0:
                return self.precpred(self._ctx, 3)
         

            if predIndex == 1:
                return self.precpred(self._ctx, 2)
         

            if predIndex == 2:
                return self.precpred(self._ctx, 4)
         




